# tool functions to compute convolutions between PDFs and coefficients functions but also between matching functions and
# coefficient functions and between matching functions alone

import numpy as np
import scipy.integrate as integrate

from .light_tools import apply_isospin_roation


def PDFConvolute(func1, pdf, x, Q, p1, nf, pid=None, target_dict=None):
    lower = x
    upper = 1.0
    if pid == 21:
        result, error = integrate.quad(
            lambda z: func1(z, Q, p1, nf) * pdf.xfxQ2(pid, x * (1.0 / z), Q * Q),
            lower,
            upper,
            epsabs=1e-12,
            epsrel=1e-6,
            limit=200,
            points=(x, 1.0),
        )
    elif pid in [4, 5, 6]:
        result, error = integrate.quad(
            lambda z: func1(z, Q, p1, nf)
            * (
                pdf.xfxQ2(pid, x * (1.0 / z), Q * Q)
                + pdf.xfxQ2(-pid, x * (1.0 / z), Q * Q)
            ),
            lower,
            upper,
            epsabs=1e-12,
            epsrel=1e-6,
            limit=200,
            points=(x, 1.0),
        )
    else:

        def light_pdfs(z):
            light_f = [1, 2, 3]
            if nf > 4:
                light_f.extend([4])
            if nf > 5:
                light_f.extend([5])
            light_pdfs = apply_isospin_roation(pdf, x / z, Q, light_f, target_dict)
            return np.sum(light_pdfs)

        result, error = integrate.quad(
            lambda z: func1(z, Q, p1, nf) * light_pdfs(z),
            lower,
            upper,
            epsabs=1e-12,
            epsrel=1e-6,
            limit=200,
            points=(x, 1.0),
        )
    return result


def Convolute(func1, matching, x, Q, p1, nf, nl=None):
    lower = x
    upper = 1.0
    nl = nf if nl is None else nl
    result, _ = integrate.quad(
        lambda z: (1.0 / z) * func1(z, Q, p1, nl) * matching(x * (1.0 / z), p1, nf),
        lower,
        upper,
        epsabs=1e-12,
        epsrel=1e-6,
        limit=200,
        points=(x, 1.0),
    )
    return result


def Convolute_matching(matching1, matching2, x, Q, p1, nf):
    lower = x
    upper = 1.0
    result, error = integrate.quad(
        lambda z: (1.0 / z) * matching1(z, p1, nf) * matching2(x * (1.0 / z), p1, nf),
        lower,
        upper,
        epsabs=1e-12,
        epsrel=1e-6,
        limit=200,
        points=(x, 1.0),
    )
    return result


def PDFConvolute_plus(func1, pdf, x, Q, p1, nf, pid=None, target_dict=None):
    if pid == 21:
        plus1, error1 = integrate.quad(
            lambda z: func1(z, Q, p1, nf)
            * (pdf.xfxQ2(pid, x * (1.0 / z), Q * Q) - pdf.xfxQ2(pid, x, Q * Q)),
            x,
            1.0,
            epsabs=1e-12,
            epsrel=1e-6,
            limit=200,
            points=(x, 1.0),
        )
    elif pid in [4, 5, 6]:
        plus1, error1 = integrate.quad(
            lambda z: func1(z, Q, p1, nf)
            * (
                pdf.xfxQ2(pid, x * (1.0 / z), Q * Q)
                + pdf.xfxQ2(-pid, x * (1.0 / z), Q * Q)
                - pdf.xfxQ2(pid, x, Q * Q)
                - pdf.xfxQ2(-pid, x, Q * Q)
            ),
            x,
            1.0,
            epsabs=1e-12,
            epsrel=1e-6,
            limit=200,
            points=(x, 1.0),
        )
    else:

        def light_pdfs(z):
            light_f = [1, 2, 3]
            if nf > 4:
                light_f.extend([4])
            if nf > 5:
                light_f.extend([5])
            light_pdfs = apply_isospin_roation(pdf, x / z, Q, light_f, target_dict)
            light_pdfs_x = apply_isospin_roation(pdf, x, Q, light_f, target_dict)
            return np.sum(light_pdfs - light_pdfs_x)

        plus1, error1 = integrate.quad(
            lambda z: func1(z, Q, p1, nf) * light_pdfs(z),
            x,
            1.0,
            epsabs=1e-12,
            epsrel=1e-6,
            limit=200,
            points=(x, 1.0),
        )
    return plus1


def Convolute_plus_coeff(func1, matching, x, Q, p1, nf):
    plus1, error1 = integrate.quad(
        lambda z: func1(z, Q, p1, nf)
        * ((1.0 / z) * matching(x * (1.0 / z), p1, nf) - matching(x, p1, nf)),
        x,
        1.0,
        epsabs=1e-12,
        epsrel=1e-6,
        limit=200,
        points=(x, 1.0),
    )
    return plus1


def Convolute_plus_matching(func1, matching, x, Q, p1, nf, nl=None):
    nl = nf if nl is None else nl
    plus1, error1 = integrate.quad(
        lambda z: matching(z, p1, nf)
        * ((1.0 / z) * func1(x * (1.0 / z), Q, p1, nl) - func1(x, Q, p1, nf)),
        x,
        1.0,
        epsabs=1e-12,
        epsrel=1e-6,
        limit=200,
        points=(x, 1.0),
    )
    return plus1



import numpy as np
import pandas as pd
import yaml
from eko.couplings import Couplings
from eko.io import types
from eko.thresholds import ThresholdsAtlas

from . import parameters
from .logging import console


class TheoryParameters:
    """Class containing all the theory parameters."""

    def __init__(self, order, fns, masses, sc, thr_atlas, grids, full_card=None):
        self.order = order
        self.fns = fns
        self.masses = masses
        self.grids = grids
        self._t_card = full_card
        self.strong_coupling = sc
        self.thr_atlas = thr_atlas
        self.n3lo_variation = full_card.get("n3lo_cf_variation", 0)

        if not self.grids:
            console.log(
                "[yellow underline]Warning, grids are not enabled, this might take a while ..."
            )

    def yadism_like(self):
        return self._t_card

    @classmethod
    def load_card(cls, configs, name):
        """Return a TheoryParameters object."""
        if isinstance(name, str):
            with open(
                configs["paths"]["theory_cards"] / (name + ".yaml"), encoding="utf-8"
            ) as f:
                th = yaml.safe_load(f)
        else:
            th = name

        # Disable some NNPDF features not included here
        if "TMC" in th and th["TMC"] != 0:
            console.log(
                "[red underline]Warning, disable Target Mass Corrections:", "TMC=0"
            )
            th["TMC"] = 0
        if "IC" in th and th["IC"] == 1:
            console.log("[red underline]Warning, disable Intrinsic Charm:", "IC=0")
            th["IC"] = 0
        if "FactScaleVar" in th and th["FactScaleVar"]:
            console.log(
                "[red underline]Warning, disable Factorization Scale Variation:",
                "FactScaleVar=0",
            )
            th["FactScaleVar"] = False
        if "RenScaleVar" in th and th["RenScaleVar"]:
            console.log(
                "[red underline]Warning, disable Renormalization Scale Variation:",
                "RenScaleVar=0",
            )
            th["RenScaleVar"] = False

        # compatibility layer
        # PTO
        if "order" in th:
            order = th["order"]
        else:
            order = th["PTO"]

        fns = th.get("fns", "fonll")
        grids = th.get("grids", True)
        mc = th.get("mc", parameters.default_masses(4))
        mb = th.get("mb", parameters.default_masses(5))
        mt = th.get("mt", parameters.default_masses(6))
        masses = np.array([mc, mb, mt])
        kmc = th.get("kcThr", 1.0)
        kmb = th.get("kbThr", 1.0)
        kmt = th.get("ktThr", 1.0)
        thresholds_ratios = np.array([kmc, kmb, kmt]) ** 2
        method = types.CouplingEvolutionMethod.EXPANDED
        if "ModEv" in th and th["ModEv"] == "EXA":
            method = types.CouplingEvolutionMethod.EXACT

        ref = types.CouplingsRef(
            alphas=types.FloatRef(
                value=th.get("alpahs", 0.118), scale=th.get("Qref", 91.2)
            ),
            alphaem=types.FloatRef(value=th.get("alphaqed", 0.007496252), scale=np.nan),
            max_num_flavs=th.get("MaxNfAs", 5),
            num_flavs_ref=th.get("nfref", 5),
        )
        sc = Couplings(
            couplings=ref,
            order=(order + 1, 0),
            method=method,
            masses=masses**2,
            hqm_scheme=types.QuarkMassSchemes.POLE,
            thresholds_ratios=thresholds_ratios,
        )
        thr_atlas = ThresholdsAtlas(
            masses=masses**2, thresholds_ratios=thresholds_ratios
        )
        return cls(
            order=order,
            fns=fns,
            grids=grids,
            masses=masses,
            sc=sc,
            thr_atlas=thr_atlas,
            full_card=th,
        )


class Observable:
    """Class describing observable settings"""

    def __init__(self, name, heavyness, pdf, restype, kinematics):
        self.name = name
        self.pdf = pdf
        self.restype = restype
        self.heavyness = heavyness
        self.kinematics = pd.DataFrame(kinematics)

    @property
    def x_grid(self):
        return self.kinematics.x.values

    @property
    def q_grid(self):
        return self.kinematics.q.values

    @property
    def y_grid(self):
        return self.kinematics.y.values


class OperatorParameters:
    """Class containing all the operator parameters."""

    def __init__(self, obs, name, target_dict, full_card=None):
        self.obs = obs
        self._o_card = full_card
        self.dataset_name = name
        self.target_dict = target_dict

    def yadism_like(self):
        return self._o_card

    @classmethod
    def load_card(cls, configs, name, pdf_name=None):
        """Return a OperatorParameters object."""
        if isinstance(name, str):
            with open(
                configs["paths"]["operator_cards"] / (name + ".yaml"), encoding="utf-8"
            ) as f:
                obs = yaml.safe_load(f)
        else:
            obs = name

        # Disables some NNPDF settings
        # TODO: implement NC and F3
        if "prDIS" in obs and obs["prDIS"] != "EM":
            console.log("[red underline]Warning, setting:", "prDIS = EM")
            obs["prDIS"] = "EM"
        if "ProjectileDIS" in obs and obs["ProjectileDIS"] not in [
            "electron",
            "positron",
        ]:
            console.log("[red underline]Warning, setting ProjectileDIS = electron")
            obs["ProjectileDIS"] = "electron"

        target_dict = {"A": 1, "Z": 1}
        if "TargetDIS" in obs:
            if obs["TargetDIS"] not in ["proton", "isoscalar"]:
                raise NotImplementedError(f"{obs['TargetDIS']} is not available")
            if obs["TargetDIS"] == "isoscalar":
                target_dict = {"A": 2, "Z": 1}

        # DIS_TP runcards
        observables = []
        if "obs" in obs:
            observables = []
            for ob in obs["obs"]:
                # TODO: light and total are always in FONLL mode
                # here heavyness and resytpe are not yet disentangled:
                # restype sholud come from the theory runcard
                heavyness = ob.split("_")[1]
                restype = obs["obs"][ob]["restype"]
                if heavyness in ["light", "total"]:
                    if "_incomplete" in restype and heavyness == "total":
                        restype = "total_incomplete"
                    else:
                        restype = heavyness

                observables.append(
                    Observable(
                        name=ob.split("_")[0],
                        heavyness=heavyness,
                        pdf=obs["obs"][ob]["PDF"],
                        restype=restype,
                        kinematics=obs["obs"][ob]["kinematics"],
                    )
                )
        # Yadism runcard
        else:
            for fx, kins in obs["observables"].items():
                new_kins = [
                    {"x": point["x"], "q": np.sqrt(point["Q2"]), "y": point["y"]}
                    for point in kins
                ]

                # whenever you are running a Kfact only FONLL is allowed
                restype = "FONLL"
                try:
                    heavyness = fx.split("_")[1]
                except IndexError:
                    heavyness = "total"
                if heavyness in ["light", "total"]:
                    restype = heavyness

                observables.append(
                    Observable(
                        name=fx.split("_")[0],
                        heavyness=heavyness,
                        pdf=pdf_name,
                        restype=restype,
                        kinematics=new_kins,
                    )
                )
        return cls(observables, name, target_dict, full_card=obs)


class RunParameters:
    """Class to hold all the running parameters."""

    def __init__(self, theoryparam, operatorparam, resultpath):
        self.theoryparam = theoryparam
        self.operatorparam = operatorparam
        self.resultpath = resultpath
        self.results = {}

    def theory_parameters(self):
        return self.theoryparam

    def operator_parameters(self):
        return self.operatorparam

    def resultpath(self):
        return self.resultpath

    def dump_results(self):
        for ob, res in self.results.items():
            self.dump_result(ob, res)

    def dump_result(self, ob, ob_result):
        heavyness_dict = {"charm": ["4", 0], "bottom": ["5", 1]}
        thr_ratio = np.sqrt(
            self.theoryparam.thr_atlas.thresholds_ratios[
                heavyness_dict[ob.heavyness][1]
            ]
        )
        file_name = (
            ob.name
            + "_"
            + ob.restype
            + "_"
            + str(self.theory_parameters().order)
            + "_"
            + heavyness_dict[ob.heavyness][0]
            + "_"
            + ob.heavyness
            + "_thr="
            + str(thr_ratio)
            + "_"
            + str(ob.pdf)
            + "_"
            + str(self.theory_parameters().n3lo_variation)
        )
        obs_path = self.resultpath / (file_name + ".yaml")
        # construct the object to dump
        to_dump = dict(
            x_grid=ob.x_grid.tolist(),
            q_grid=ob.q_grid.tolist(),
            obs=ob_result.tolist(),
        )
        console.log(f"[green]Saving results for {ob.name} in {obs_path}")
        with open(obs_path, "w", encoding="UTF-8") as f:
            yaml.safe_dump(to_dump, f)



"""FL structure function"""
import numpy as np

from .. import MassiveCoeffFunc, MasslessCoeffFunc, TildeCoeffFunc, TildeCoeffFunc_light
from ..parameters import (
    alpha_s,
    charges,
    masses,
    n3lo_color_factors,
    number_active_flavors,
    number_light_flavors,
    pids,
)
from .heavy_tools import PDFConvolute
from .light_tools import (
    PDFConvolute_light,
    PDFConvolute_light_singlet,
    mkPDF,
    non_singlet_pdf,
)

g_id = pids["g"]


def FL_FO(
    order, pdf, x, Q, h_id, meth=None, muF_ratio=1, target_dict=None, muR_ratio=1
):
    """
    Compute the FO results for the structure function FL

    Parameters:
        order : int
            requested perturbative order (0 == LO, 1 == NLO,...)
        pdf : str or list(str)
            pdf(s) to be used
        x : float
            x-value
        Q : float
            Q-value
        h_id : int
            heavy quark id
        muF_ratio : float
            ratio to Q of the factorization scale
        muR_ratio : float
            ratio to Q of the renormalization scale
    Returns:
            : float
            result
    """
    Mypdf = mkPDF(pdf, order)
    muR = muR_ratio * Q
    p = [masses(h_id), Q, charges(h_id)]
    nf = number_active_flavors(Q)
    conv_func = PDFConvolute
    if nf != h_id:
        conv_func = PDFConvolute_light_singlet
    a_s = alpha_s(muR**2, Q**2)
    if order >= 0:
        res = 0.0
    if order >= 1:
        res += a_s * PDFConvolute(
            MassiveCoeffFunc.CLg_1_m_reg, Mypdf, x, Q, p, h_id, g_id
        )
    if order >= 2:
        res += a_s**2 * (
            PDFConvolute(MassiveCoeffFunc.CLg_2_m_reg, Mypdf, x, Q, p, h_id, g_id)
            + conv_func(
                MassiveCoeffFunc.CLq_2_m_reg,
                Mypdf,
                x,
                Q,
                p,
                h_id,
                target_dict=target_dict,
            )
        )

        # add missing diagrams with hq+1 effects
        for ihq in range(h_id + 1, 6):
            pihq = [masses(ihq), Q, charges(h_id)]
            reg_miss = PDFConvolute(
                MassiveCoeffFunc.CLb_2_m_reg, Mypdf, x, Q, pihq, h_id, h_id
            )
            res += a_s**2 * reg_miss

    if order >= 3:
        res += a_s**3 * (
            PDFConvolute(MassiveCoeffFunc.CLg_3_m_reg, Mypdf, x, Q, p, h_id, g_id)
            + conv_func(
                MassiveCoeffFunc.CLq_3_m_reg,
                Mypdf,
                x,
                Q,
                p,
                h_id,
                target_dict=target_dict,
            )
        )
    return res


def FL_R(order, pdf, x, Q, h_id, meth=None, muF_ratio=1, target_dict=None, muR_ratio=1):
    """
    Compute the R result for the structure function FL

    Parameters:
        order : int
            requested perturbative order (0 == LO, 1 == NLO,...)
        pdf : str or list(str)
            pdf(s) to be used
        x : float
            x-value
        Q : float
            Q-value
        h_id : int
            heavy quark id
        muF_ratio : float
            ratio to Q of the factorization scale
        muR_ratio : float
            ratio to Q of the renormalization scale
    Returns:
            : float
            result
    """
    Mypdf = mkPDF(pdf, order)
    muR = muR_ratio * Q
    p = [masses(h_id), Q, charges(h_id)]
    nf = number_active_flavors(Q)
    a_s = alpha_s(muR**2, Q**2)
    res = 0.0
    if order >= 0:
        res = 0.0
    if order >= 1:
        res += a_s * PDFConvolute(MasslessCoeffFunc.CLg_1_reg, Mypdf, x, Q, p, nf, g_id)
    if order >= 2:
        nnll_reg = a_s * (
            a_s
            * (
                PDFConvolute(MasslessCoeffFunc.CLg_2_reg, Mypdf, x, Q, p, nf, g_id)
                + PDFConvolute(
                    MasslessCoeffFunc.CLq_2_reg,
                    Mypdf,
                    x,
                    Q,
                    p,
                    nf,
                    target_dict=target_dict,
                )
            )
            + PDFConvolute(MasslessCoeffFunc.CLb_1_reg, Mypdf, x, Q, p, nf, h_id)
        )
        res += nnll_reg
    if order >= 3:
        pg = ps = [masses(h_id), Q, 0, charges(h_id)]
        ps[2] = n3lo_color_factors("s", nf, False)
        pg[2] = n3lo_color_factors("g", nf, False)
        n3ll_reg = a_s**2 * (
            a_s
            * (
                PDFConvolute(MasslessCoeffFunc.CLg_3_reg, Mypdf, x, Q, pg, nf, g_id)
                + PDFConvolute(
                    MasslessCoeffFunc.CLq_3_reg,
                    Mypdf,
                    x,
                    Q,
                    ps,
                    nf,
                    target_dict=target_dict,
                )
            )
            + PDFConvolute(MasslessCoeffFunc.CLb_2_reg, Mypdf, x, Q, p, nf, h_id)
        )
        n3ll_loc = a_s**2 * (
            MasslessCoeffFunc.CLb_2_loc(x, Q, p, nf)
            * (Mypdf.xfxQ2(h_id, x, Q * Q) + Mypdf.xfxQ2(-h_id, x, Q * Q))
        )
        res += n3ll_reg + n3ll_loc
    return res


def FL_M(order, pdf, x, Q, h_id, meth, muF_ratio=1, target_dict=None, muR_ratio=1):
    """
    Compute the M result for the structure function FL

    Parameters:
        order : int
            requested perturbative order (0 == LO, 1 == NLO,...)
        meth : str
            method to be used (our, fonll)
        pdf : str or list(str)
            pdf(s) to be used
        x : float
            x-value
        Q : float
            Q-value
        h_id : int
            heavy quark id
        muF_ratio : float
            ratio to Q of the factorization scale
        muR_ratio : float
            ratio to Q of the renormalization scale
    Returns:
            : float
            result
    """
    Mypdf = mkPDF(pdf, order)
    muR = muR_ratio * Q
    nf = number_active_flavors(Q)
    p = [masses(h_id), Q, charges(h_id)]
    a_s = alpha_s(muR**2, Q**2)
    if meth == "our":
        if order >= 0:
            res = 0.0
        if order >= 1:
            res += a_s * PDFConvolute(
                TildeCoeffFunc.CLg_1_til_reg, Mypdf, x, Q, p, nf, g_id
            )
        if order >= 2:
            nnlo_nnll_reg = a_s * (
                a_s
                * (
                    PDFConvolute(TildeCoeffFunc.CLg_2_til_reg, Mypdf, x, Q, p, nf, g_id)
                    + PDFConvolute(
                        TildeCoeffFunc.CLq_2_til_reg,
                        Mypdf,
                        x,
                        Q,
                        p,
                        nf,
                        target_dict=target_dict,
                    )
                )
                + PDFConvolute(MasslessCoeffFunc.CLb_1_reg, Mypdf, x, Q, p, nf, h_id)
            )
            res += nnlo_nnll_reg
        if order >= 3:
            n3lo_n3ll_reg = a_s**2 * (
                a_s
                * (
                    PDFConvolute(TildeCoeffFunc.CLg_3_til_reg, Mypdf, x, Q, p, nf, g_id)
                    + PDFConvolute(
                        TildeCoeffFunc.CLq_3_til_reg,
                        Mypdf,
                        x,
                        Q,
                        p,
                        nf,
                        target_dict=target_dict,
                    )
                )
                + PDFConvolute(MasslessCoeffFunc.CLb_2_reg, Mypdf, x, Q, p, nf, h_id)
            )
            n3lo_n3ll_loc = (
                a_s
                * a_s
                * MasslessCoeffFunc.CLb_2_loc(x, Q, p, nf)
                * (Mypdf.xfxQ2(h_id, x, Q * Q) + Mypdf.xfxQ2(-h_id, x, Q * Q))
            )
            res += n3lo_n3ll_loc + n3lo_n3ll_reg
    if meth == "fonll":
        if order >= 0:
            res = 0.0
        if order >= 1:
            res += a_s * (
                PDFConvolute(TildeCoeffFunc.CLg_1_til_reg, Mypdf, x, Q, p, nf, g_id)
                + PDFConvolute(MasslessCoeffFunc.CLb_1_reg, Mypdf, x, Q, p, nf, h_id)
            )
        if order >= 2:
            nnlo_nnll_reg = a_s**2 * (
                PDFConvolute(TildeCoeffFunc.CLg_2_til_reg, Mypdf, x, Q, p, nf, g_id)
                + PDFConvolute(
                    TildeCoeffFunc.CLq_2_til_reg,
                    Mypdf,
                    x,
                    Q,
                    p,
                    nf,
                    target_dict=target_dict,
                )
                + PDFConvolute(MasslessCoeffFunc.CLb_2_reg, Mypdf, x, Q, p, nf, h_id)
            )
            nnlo_nnll_loc = (
                a_s**2
                * MasslessCoeffFunc.CLb_2_loc(x, Q, p, nf)
                * (Mypdf.xfxQ2(h_id, x, Q * Q) + Mypdf.xfxQ2(-h_id, x, Q * Q))
            )
            res += nnlo_nnll_reg + nnlo_nnll_loc

            # add missing diagrams with hq+1 effects
            for ihq in range(h_id + 1, 6):
                pihq = [masses(ihq), Q, charges(h_id)]
                reg_miss = PDFConvolute(
                    MassiveCoeffFunc.CLb_2_m_reg, Mypdf, x, Q, pihq, nf, h_id
                )
                res += a_s**2 * reg_miss

        if order >= 3:
            pns = [masses(h_id), Q, 0, charges(h_id)]
            pns[2] = n3lo_color_factors("ns", nf, False)
            n3lo_n3ll_reg = a_s**3 * (
                PDFConvolute(TildeCoeffFunc.CLg_3_til_reg, Mypdf, x, Q, p, nf, g_id)
                + PDFConvolute(
                    TildeCoeffFunc.CLq_3_til_reg,
                    Mypdf,
                    x,
                    Q,
                    p,
                    nf,
                    target_dict=target_dict,
                )
                + PDFConvolute(MasslessCoeffFunc.CLb_3_reg, Mypdf, x, Q, pns, nf, h_id)
            )
            n3lo_n3ll_loc = (
                a_s**3
                * MasslessCoeffFunc.CLb_3_loc(x, Q, pns, nf)
                * (Mypdf.xfxQ2(h_id, x, Q * Q) + Mypdf.xfxQ2(-h_id, x, Q * Q))
            )
            res += n3lo_n3ll_reg + n3lo_n3ll_loc
    return res


def FL_Light(order, pdf, x, Q, h_id=None, meth=None, target_dict=None, muR_ratio=1):
    """
    Compute the light contribution for the structure function FL

    Parameters:
        order : int
            requested perturbative order (0 == LO, 1 == NLO,...)
        pdf : str or list(str)
            pdf(s) to be used
        x : float
            x-value
        Q : float
            Q-value
        h_id : int
            heavy quark id, None
        muR_ratio : float
            ratio to Q of the renormalization scale
    Returns:
            : float
            result
    """
    Mypdf = mkPDF(pdf, order)
    muR = muR_ratio * Q
    p = [0, Q, 1]
    nl = number_light_flavors(Q)
    a_s = alpha_s(muR**2, Q**2)
    meansq_e = np.mean([charges(nl) ** 2 for nl in range(1, nl + 1)])
    if order >= 0:
        res = 0.0
    if order >= 1:
        reg = PDFConvolute_light(
            MasslessCoeffFunc.CLb_1_reg, Mypdf, x, Q, p, nl, target_dict
        ) + nl * meansq_e * PDFConvolute(
            MasslessCoeffFunc.CLg_1_reg, Mypdf, x, Q, p, nl, g_id
        )
        res += a_s * reg
    if order >= 2:
        if nl != number_active_flavors(Q):
            p = [masses(nl + 1), Q, 1]
            reg = PDFConvolute_light(
                TildeCoeffFunc_light.CLb_2_til_reg, Mypdf, x, Q, p, nl, target_dict
            ) + nl * meansq_e * (
                PDFConvolute(
                    TildeCoeffFunc_light.CLg_2_til_reg, Mypdf, x, Q, p, nl, g_id
                )
                + PDFConvolute_light_singlet(
                    TildeCoeffFunc_light.CLq_2_til_reg, Mypdf, x, Q, p, nl, target_dict
                )
            )
            loc = TildeCoeffFunc_light.CLb_2_til_loc(x, Q, p, nl) * non_singlet_pdf(
                Mypdf, x, Q, nl, target_dict
            )
            res += a_s**2 * (reg + loc)

            # Add heavy quark contribution pure singlet contribution,
            # note the CF has to be evalueted with nl+1 here, only the
            # heavy quark coupling is set to 0
            singlet_h = (
                +nl
                * meansq_e
                * (
                    +PDFConvolute(
                        MasslessCoeffFunc.CLq_2_reg, Mypdf, x, Q, p, nl + 1, nl + 1
                    )
                )
            )
            res += a_s**2 * singlet_h
        else:
            reg = PDFConvolute_light(
                MasslessCoeffFunc.CLb_2_reg, Mypdf, x, Q, p, nl, target_dict
            ) + nl * meansq_e * (
                PDFConvolute(MasslessCoeffFunc.CLg_2_reg, Mypdf, x, Q, p, nl, g_id)
                + PDFConvolute_light_singlet(
                    MasslessCoeffFunc.CLq_2_reg, Mypdf, x, Q, p, nl, target_dict
                )
            )
            loc = MasslessCoeffFunc.CLb_2_loc(x, Q, p, nl) * non_singlet_pdf(
                Mypdf, x, Q, nl, target_dict
            )
            res += a_s**2 * (reg + loc)

        # add the missing terms for heavy quarks
        for ihq in range(nl + 1, 6):
            pihq = [masses(ihq), Q, 1]
            nf = number_active_flavors(Q)

            # for the thr quark we subtract the asymptotic
            if ihq == nl + 1 and nl != nf:
                reg_miss = PDFConvolute_light(
                    MassiveCoeffFunc.CLb_2_m_reg, Mypdf, x, Q, pihq, nl, target_dict
                )
                res += a_s**2 * reg_miss
                reg_asy = PDFConvolute_light(
                    TildeCoeffFunc_light.CLb_2_asy_reg,
                    Mypdf,
                    x,
                    Q,
                    pihq,
                    nl,
                    target_dict,
                )
                res -= a_s**2 * reg_asy

            else:
                reg_miss = PDFConvolute_light(
                    MassiveCoeffFunc.CLb_2_m_reg, Mypdf, x, Q, pihq, nf, target_dict
                )
                res += a_s**2 * reg_miss

    if order >= 3:
        pg = ps = pns = [0, Q, 0, 1]
        if nl != number_active_flavors(Q):
            # NOTE: here the NS has to be evaluated at nl+1 but convluted with nl
            pns[2] = n3lo_color_factors("ns", nl + 1, True)
            ps[2] = n3lo_color_factors("s", nl + 1, True)
            pg[2] = n3lo_color_factors("g", nl + 1, True)
            reg = PDFConvolute_light(
                TildeCoeffFunc_light.CLb_3_til_reg, Mypdf, x, Q, pns, nl, target_dict
            ) + nl * meansq_e * (
                PDFConvolute(
                    TildeCoeffFunc_light.CLg_3_til_reg, Mypdf, x, Q, pg, nl, g_id
                )
                + PDFConvolute_light_singlet(
                    TildeCoeffFunc_light.CLq_3_til_reg, Mypdf, x, Q, ps, nl, target_dict
                )
            )
            loc = TildeCoeffFunc_light.CLb_3_til_loc(x, Q, pns, nl) * non_singlet_pdf(
                Mypdf, x, Q, nl, target_dict
            )
            res += a_s**3 * (reg + loc)

            # here we can only add the Singlet contribution from heavy quark
            ps[2] = n3lo_color_factors("s", nl + 1, False)
            singlet_h = (
                +nl
                * meansq_e
                * (
                    PDFConvolute(
                        MasslessCoeffFunc.CLq_3_reg, Mypdf, x, Q, ps, nl + 1, nl + 1
                    )
                )
            )
            res += a_s**3 * singlet_h

        else:
            pns[2] = n3lo_color_factors("ns", nl, False)
            ps[2] = n3lo_color_factors("s", nl, False)
            pg[2] = n3lo_color_factors("g", nl, False)
            reg = PDFConvolute_light(
                MasslessCoeffFunc.CLb_3_reg, Mypdf, x, Q, pns, nl, target_dict
            ) + nl * meansq_e * (
                PDFConvolute(MasslessCoeffFunc.CLg_3_reg, Mypdf, x, Q, pg, nl, g_id)
                + PDFConvolute_light_singlet(
                    MasslessCoeffFunc.CLq_3_reg, Mypdf, x, Q, ps, nl, target_dict
                )
            )
            loc = MasslessCoeffFunc.CLb_3_loc(x, Q, pns, nl) * non_singlet_pdf(
                Mypdf, x, Q, nl, target_dict
            )
            res += a_s**3 * (reg + loc)

    return res


def FL_ZM(
    order, pdf, x, Q, h_id, meth=None, target_dict=None, muR_ratio=1, min_order=0
):
    """
    Compute the ZM heavy contribution to structure function FL

    Parameters:
        order : int
            requested perturbative order (0 == LO, 1 == NLO,...)
        pdf : str or list(str)
            pdf(s) to be used
        x : float
            x-value
        Q : float
            Q-value
        h_id : int
            heavy quark id
        muR_ratio : float
            ratio to Q of the renormalization scale
    Returns:
            : float
            result
    """
    Mypdf = mkPDF(pdf, order)
    muR = muR_ratio * Q
    nl = number_active_flavors(Q)
    conv_func = PDFConvolute
    if nl != h_id:
        conv_func = PDFConvolute_light_singlet
    p = [0, Q, charges(h_id)]
    a_s = alpha_s(muR**2, Q**2)
    pdfxfx = Mypdf.xfxQ2(h_id, x, Q**2) + Mypdf.xfxQ2(-h_id, x, Q**2)
    res = 0
    if order >= 1 and min_order <= 1:
        reg = PDFConvolute(
            MasslessCoeffFunc.CLb_1_reg, Mypdf, x, Q, p, nl, h_id
        ) + PDFConvolute(MasslessCoeffFunc.CLg_1_reg, Mypdf, x, Q, p, nl, g_id)
        res += a_s * reg
    if order >= 2 and min_order <= 2:
        reg = (
            PDFConvolute(MasslessCoeffFunc.CLb_2_reg, Mypdf, x, Q, p, nl, h_id)
            + PDFConvolute(MasslessCoeffFunc.CLg_2_reg, Mypdf, x, Q, p, nl, g_id)
            + conv_func(
                MasslessCoeffFunc.CLq_2_reg, Mypdf, x, Q, p, nl, target_dict=target_dict
            )
        )
        loc = MasslessCoeffFunc.CLb_2_loc(x, Q, p, nl) * pdfxfx
        res += a_s**2 * (reg + loc)
    if order >= 3 and min_order <= 3:
        pg = ps = pns = [0, Q, 0, charges(h_id)]
        pns[2] = n3lo_color_factors("ns", nl, False)
        ps[2] = n3lo_color_factors("s", nl, False)
        pg[2] = n3lo_color_factors("g", nl, False)
        reg = (
            PDFConvolute(MasslessCoeffFunc.CLb_3_reg, Mypdf, x, Q, pns, nl, h_id)
            + PDFConvolute(MasslessCoeffFunc.CLg_3_reg, Mypdf, x, Q, pg, nl, g_id)
            + conv_func(
                MasslessCoeffFunc.CLq_3_reg,
                Mypdf,
                x,
                Q,
                ps,
                nl,
                target_dict=target_dict,
            )
        )
        loc = MasslessCoeffFunc.CLb_3_loc(x, Q, pns, nl) * pdfxfx
        res += a_s**3 * (reg + loc)
    return res


def FL_FONLL(order, pdf, x, Q, h_id, meth, target_dict=None, muR_ratio=1):
    """
    Compute the Yadism like FONLL structure function FL

    Parameters:
        order : int
            requested perturbative order (0 == LO, 1 == NLO,...)
        pdf : str or list(str)
            pdf(s) to be used
        x : float
            x-value
        Q : float
            Q-value
        h_id : int
            heavy quark id
        muR_ratio : float
            ratio to Q of the renormalization scale
    Returns:
            : float
            result
    """
    nf = number_active_flavors(Q)
    if nf <= h_id - 1:
        return FL_FO(
            order, pdf, x, Q, h_id, target_dict=target_dict, muR_ratio=muR_ratio
        )
    elif nf == h_id:
        return FL_M(
            order, pdf, x, Q, h_id, meth, target_dict=target_dict, muR_ratio=muR_ratio
        )
    elif nf >= h_id + 1:
        return FL_ZM(
            order, pdf, x, Q, h_id, target_dict=target_dict, muR_ratio=muR_ratio
        )


def FL_Total(order, pdf, x, Q, h_id, meth, target_dict=None, muR_ratio=1):
    """
    Compute the total structure function FL.

    Parameters:
        order : int
            requested perturbative order (0 == LO, 1 == NLO,...)
        meth : str
            method to be used (our, fonll)
        pdf : str or list(str)
            pdf(s) to be used
        x : float
            x-value
        Q : float
            Q-value
        muR_ratio : float
            ratio to Q of the renormalization scale
    Returns:
        : float
            result
    """
    nf = number_active_flavors(Q)
    if nf <= 4:
        res = (
            FL_Light(
                order, pdf, x, Q, 3, meth, target_dict=target_dict, muR_ratio=muR_ratio
            )
            + FL_FONLL(
                order, pdf, x, Q, 4, meth, target_dict=target_dict, muR_ratio=muR_ratio
            )
            + FL_FO(
                order, pdf, x, Q, 5, meth, target_dict=target_dict, muR_ratio=muR_ratio
            )
        )
    elif nf == 5:
        res = FL_Light(
            order, pdf, x, Q, 4, meth, target_dict=target_dict, muR_ratio=muR_ratio
        ) + FL_FONLL(
            order, pdf, x, Q, 5, meth, target_dict=target_dict, muR_ratio=muR_ratio
        )
    else:
        res = FL_Light(
            order, pdf, x, Q, 5, meth, target_dict=target_dict, muR_ratio=muR_ratio
        )
    return res


def FL_FONLL_incomplete(order, pdf, x, Q, h_id, meth, target_dict=None, muR_ratio=1):
    r"""
    Compute the incomplete FONLL = \sum_{i=0}^{order-1} FONLL_{i} + ZM_{order}

    Parameters:
        order : int
            requested perturbative order (0 == LO, 1 == NLO,...)
        pdf : str or list(str)
            pdf(s) to be used
        x : float
            x-value
        Q : float
            Q-value
        h_id : int
            heavy quark id
        muR_ratio : float
            ratio to Q of the renormalization scale
    Returns:
            : float
            result
    """
    nf = number_active_flavors(Q)
    if nf <= h_id - 1:
        return FL_FO(
            order - 1, pdf, x, Q, h_id, target_dict=target_dict, muR_ratio=muR_ratio
        )
    if h_id == nf:
        return FL_M(
            order - 1,
            pdf,
            x,
            Q,
            h_id,
            meth,
            target_dict=target_dict,
            muR_ratio=muR_ratio,
        ) + FL_ZM(
            order,
            pdf,
            x,
            Q,
            h_id,
            target_dict=target_dict,
            muR_ratio=muR_ratio,
            min_order=order,
        )
    elif nf >= h_id + 1:
        return FL_ZM(
            order, pdf, x, Q, h_id, target_dict=target_dict, muR_ratio=muR_ratio
        )


def FL_Total_incomplete(order, pdf, x, Q, h_id, meth, target_dict=None, muR_ratio=1):
    """
    Compute the total structure function FL.

    Parameters:
        order : int
            requested perturbative order (0 == LO, 1 == NLO,...)
        meth : str
            method to be used (our, fonll)
        pdf : str or list(str)
            pdf(s) to be used
        x : float
            x-value
        Q : float
            Q-value
        muR_ratio : float
            ratio to Q of the renormalization scale
    Returns:
        : float
            result
    """
    nf = number_active_flavors(Q)
    if nf <= 4:
        res = (
            FL_Light(
                order, pdf, x, Q, 3, meth, target_dict=target_dict, muR_ratio=muR_ratio
            )
            + FL_FONLL_incomplete(
                order, pdf, x, Q, 4, meth, target_dict=target_dict, muR_ratio=muR_ratio
            )
            + FL_FO(
                order - 1,
                pdf,
                x,
                Q,
                5,
                meth,
                target_dict=target_dict,
                muR_ratio=muR_ratio,
            )
        )
    elif nf == 5:
        res = FL_Light(
            order, pdf, x, Q, 4, meth, target_dict=target_dict, muR_ratio=muR_ratio
        ) + FL_FONLL_incomplete(
            order, pdf, x, Q, 5, meth, target_dict=target_dict, muR_ratio=muR_ratio
        )
    else:
        res = FL_Light(
            order, pdf, x, Q, 5, meth, target_dict=target_dict, muR_ratio=muR_ratio
        )
    return res



# This contains the tilde coefficents functions for the matched scheme.

import numpy as np
import scipy.special as special
from eko.constants import CF, TR, CA

from . import Initialize as Ini
from . import parameters
from .MassiveCoeffFunc import (
    Cg_1_m_reg,
    Cg_2_m_reg,
    Cg_3_m_reg,
    CLg_1_m_reg,
    CLg_2_m_reg,
    CLg_3_m_reg,
    CLq_2_m_reg,
    CLq_3_m_reg,
    Cq_2_m_reg,
    Cq_3_m_reg,
)
from .MasslessCoeffFunc import (
    Cb_0_loc,
    Cb_1_loc,
    Cb_1_reg,
    Cb_1_sing,
    Cb_2_loc,
    Cb_2_reg,
    Cb_2_sing,
    CLb_1_reg,
    CLb_2_loc,
    CLb_2_reg,
)
from .MatchingFunc import (
    P2,
    Mbg_1,
    Mbg_2,
    Mbg_3_reg,
    Mbq_2,
    Mbq_3_reg,
    Mgg_1_loc,
    Mgg_2_loc,
    Mgg_2_reg,
    Mgg_2_sing,
    Mgq_2_reg,
)
from .structure_functions.heavy_tools import (
    Convolute,
    Convolute_matching,
    Convolute_plus_coeff,
    Convolute_plus_matching,
)


# convolutions
def Cb1_Mbg1(z, p, _nf):
    e_h = p[-1]
    res = (
        4
        * CF
        * TR
        * e_h**2
        * (
            -(5.0 / 2.0)
            + 2 * z * (3 - 4 * z)
            + (np.pi**2 / 6.0) * (-1 + 2 * z - 4 * z**2)
            + np.log(1 - z) ** 2 * (1 - 2 * z * (1 - z))
            - (1.0 / 2.0)
            * np.log(1 - z)
            * (7 + 4 * z * (3 * z - 4) + (4 - 8 * z * (1 - z)) * np.log(z))
            + (1.0 / 2.0)
            * np.log(z)
            * (-1 + 4 * z * (3 * z - 2) + (1 - 2 * z + 4 * pow(z, 2)) * np.log(z))
            + (2 * z - 1) * special.spence(complex(1 - z))
        )
    )
    return np.real(res)


def CLb1_Mbg1(z, p, _nf):
    e_h = p[-1]
    return 8 * CF * TR * pow(e_h, 2) * (1 + z - 2 * pow(z, 2) + 2 * z * np.log(z))


def Mbg1_Mgg2_sing(x, p, _nf):
    L = np.log((p[1] ** 2) / (p[0] ** 2))
    return (
        16
        / 27
        * CA
        * L
        * (28 - 30 * L + 9 * L**2)
        * TR**2
        * (-1 + (4 - 3 * x) * x + (-1 - 2 * (-1 + x) * x) * np.log(x))
    )


# F2
def Cg_1_til_reg(z, Q, p, _nf):
    return Cg_1_m_reg(z, Q, p, _nf - 1) - 2 * Cb_0_loc(z, Q, p, _nf) * Mbg_1(z, p, _nf)


def Cg_2_til_reg(z, Q, p, _nf):
    return (
        Cg_2_m_reg(z, Q, p, _nf - 1)
        - 2
        * Cb_0_loc(z, Q, p, _nf)
        * (Mbg_2(z, p, _nf) - Mbg_1(z, p, _nf) * Mgg_1_loc(z, p, _nf))
        - 2 * np.log((Q**2) / (p[0] ** 2)) * Cb1_Mbg1(z, p, _nf)
    )


def Cg_3_til_reg(z, Q, p, nf, use_analytic=False):
    if parameters.grids and not use_analytic:
        return Ini.Cg3_til[nf - 4](z, Q)[0]
    return (
        Cg_3_m_reg(z, Q, p, nf)
        + Cg_2_m_reg(z, Q, p, nf - 1) * Mgg_1_loc(z, p, nf)
        + P2(p) * Cg_1_m_reg(z, Q, p, nf - 1)
        - (
            Cg_1_m_reg(z, Q, p, nf - 1) * Mgg_2_loc(z, p, nf)
            + Convolute(Cg_1_m_reg, Mgg_2_reg, z, Q, p, nf, nf - 1)
            + Convolute_plus_matching(Cg_1_m_reg, Mgg_2_sing, z, Q, p, nf, nf - 1)
        )
        - 2
        * Cb_0_loc(z, Q, p, nf)
        * (
            Mbg_3_reg(z, p, nf)
            - Mgg_1_loc(z, p, nf) * Mbg_2(z, p, nf)
            + Mbg_1(z, p, nf) * Mgg_1_loc(z, p, nf) * Mgg_1_loc(z, p, nf)
            - (
                Mbg_1(z, p, nf) * Mgg_2_loc(z, p, nf)
                + Convolute_matching(Mbg_1, Mgg_2_reg, z, Q, p, nf)
                + Mbg1_Mgg2_sing(z, p, nf)
            )
        )
        - 2
        * (
            Cb_1_loc(z, Q, p, nf) * Mbg_2(z, p, nf)
            + Convolute(Cb_1_reg, Mbg_2, z, Q, p, nf)
            + Convolute_plus_coeff(Cb_1_sing, Mbg_2, z, Q, p, nf)
            - Cb1_Mbg1(z, p, nf) * Mgg_1_loc(z, p, nf)
        )
        - 2
        * (
            Mbg_1(z, p, nf) * Cb_2_loc(z, Q, p, nf)
            + Convolute(Cb_2_reg, Mbg_1, z, Q, p, nf)
            + Convolute_plus_coeff(Cb_2_sing, Mbg_1, z, Q, p, nf)
        )
    )


def Cq_2_til_reg(z, Q, p, _nf):
    return Cq_2_m_reg(z, Q, p, _nf - 1) - 2 * Cb_0_loc(z, Q, p, _nf) * Mbq_2(z, p, _nf)


def Cq_3_til_reg(z, Q, p, nf, use_analytic=False):
    if parameters.grids and not use_analytic:
        return Ini.Cq3_til[nf - 4](z, Q)[0]
    return (
        Cq_3_m_reg(z, Q, p, nf)
        + 2 * Cq_2_m_reg(z, Q, p, nf - 1) * Mgg_1_loc(z, p, nf)
        - Convolute(Cg_1_m_reg, Mgq_2_reg, z, Q, p, nf, nf - 1)
        - 2
        * (
            Cb_1_loc(z, Q, p, nf) * Mbq_2(z, p, nf)
            + Convolute(Cb_1_reg, Mbq_2, z, Q, p, nf)
            + Convolute_plus_coeff(Cb_1_sing, Mbq_2, z, Q, p, nf)
        )
        - 2 * (Cb_0_loc(z, Q, p, nf) * Mbq_3_reg(z, p, nf))
    )


# FL
def CLg_1_til_reg(z, Q, p, _nf):
    return CLg_1_m_reg(z, Q, p, _nf - 1)


def CLg_2_til_reg(z, Q, p, _nf):
    return CLg_2_m_reg(z, Q, p, _nf - 1) - 2 * np.log(
        (Q**2) / (p[0] ** 2)
    ) * CLb1_Mbg1(z, p, _nf)


def CLg_3_til_reg(z, Q, p, nf, use_analytic=False):
    if parameters.grids and not use_analytic:
        return Ini.CLg3_til[nf - 4](z, Q)[0]
    return (
        CLg_3_m_reg(z, Q, p, nf)
        + CLg_2_m_reg(z, Q, p, nf - 1) * Mgg_1_loc(z, p, nf)
        + P2(p) * CLg_1_m_reg(z, Q, p, nf - 1)
        - (
            CLg_1_m_reg(z, Q, p, nf - 1) * Mgg_2_loc(z, p, nf)
            + Convolute(CLg_1_m_reg, Mgg_2_reg, z, Q, p, nf, nf - 1)
            + Convolute_plus_matching(CLg_1_m_reg, Mgg_2_sing, z, Q, p, nf, nf - 1)
        )
        - 2
        * (
            Convolute(CLb_1_reg, Mbg_2, z, Q, p, nf)
            - CLb1_Mbg1(z, p, nf) * Mgg_1_loc(z, p, nf)
        )
        - 2
        * (
            CLb_2_loc(z, Q, p, nf) * Mbg_1(z, p, nf)
            + Convolute(CLb_2_reg, Mbg_1, z, Q, p, nf)
        )
    )


def CLq_2_til_reg(z, Q, p, _nf):
    return CLq_2_m_reg(z, Q, p, _nf - 1)


def CLq_3_til_reg(z, Q, p, nf, use_analytic=False):
    if parameters.grids and not use_analytic:
        return Ini.CLq3_til[nf - 4](z, Q)[0]
    return (
        CLq_3_m_reg(z, Q, p, nf)
        + 2 * CLq_2_m_reg(z, Q, p, nf - 1) * Mgg_1_loc(z, p, nf)
        - Convolute(CLg_1_m_reg, Mgq_2_reg, z, Q, p, nf, nf - 1)
        - 2 * Convolute(CLb_1_reg, Mbq_2, z, Q, p, nf)
    )



"""Initialize the grids needed for the evaluation of coefficients functions."""
import pathlib

import numpy as np
from scipy.interpolate import interp1d, interp2d

from . import ReadTxt as readt
from .logging import console

PATH_TO_GLOBAL = str(pathlib.Path(__file__).parent.parent.parent)


def InitializeQX():
    """Initialize the Q and z grids over which the other functions are built
    from the files in External (for Niccolo functions).
    """
    global QList
    global ZList
    QList = readt.read1D(PATH_TO_GLOBAL + "/External/Q.txt")
    ZList = readt.read1D(PATH_TO_GLOBAL + "/External/x.txt")


def append_empty(grid_list, nf_list):
    """Append an empty element when only bottom is need."""
    if nf_list == [5]:
        grid_list.append(None)
    if nf_list == [6]:
        grid_list.append(None)
        grid_list.append(None)
    return grid_list


# N3LO grids
def InitializeMbg_3(nf_list):
    """Initialize the Mbg at N3LO matching condition from the file in External."""
    global Mbg3
    Mbg3 = []
    append_empty(Mbg3, nf_list)
    for nf in nf_list:
        Mbg3_array = np.array(
            readt.readND(PATH_TO_GLOBAL + f"/External/Mbg_3/Mbg3_nf{nf}.txt")
        )
        Mbg3.append(interp2d(ZList, QList, Mbg3_array.T, kind="quintic"))


def InitializeMbq_3(nf_list):
    """Initialize the Mbg at N3LO matching condition from the file in External."""
    global Mbq3
    Mbq3 = []
    append_empty(Mbq3, nf_list)
    for nf in nf_list:
        Mbq3_array = np.array(
            readt.readND(PATH_TO_GLOBAL + f"/External/Mbq_3/Mbq3_nf{nf}.txt")
        )
        Mbq3.append(interp2d(ZList, QList, Mbq3_array.T, kind="quintic"))


def InitializeCq3_m(nf_list, n3lo_variation):
    """Initialize the Cq2 at N3LO massive function from the file in External."""
    global Cq3m
    Cq3m = []
    append_empty(Cq3m, nf_list)
    for nf in nf_list:
        Cq3m_array = np.array(
            readt.readND(
                PATH_TO_GLOBAL + f"/External/Cq_3_m/C2q_nf{nf}_var{n3lo_variation}.txt"
            )
        )
        Cq3m.append(interp2d(ZList[:-1], QList, Cq3m_array[:, :-1], kind="quintic"))


def InitializeCq3_til(nf_list, n3lo_variation):
    """Initialize the Cq at N3LO tilde function from the file in External."""
    global Cq3_til
    Cq3_til = []
    append_empty(Cq3_til, nf_list)
    for nf in nf_list:
        Cq3_til_array = np.array(
            readt.readND(
                PATH_TO_GLOBAL
                + f"/External/Cq_3_til/Cq3til_nf{nf}_var{n3lo_variation}.txt"
            )
        )
        Cq3_til.append(interp2d(ZList[:-1], QList, Cq3_til_array.T, kind="quintic"))


def InitializeCLq3_m(nf_list, n3lo_variation):
    """Initialize the CqL at N3LO massive function from the file in External."""
    global CLq3m
    CLq3m = []
    append_empty(CLq3m, nf_list)
    for nf in nf_list:
        CLq3m_array = np.array(
            readt.readND(
                PATH_TO_GLOBAL + f"/External/CLq_3_m/CLq_nf{nf}_var{n3lo_variation}.txt"
            )
        )
        CLq3m.append(interp2d(ZList, QList, CLq3m_array, kind="quintic"))


def InitializeCLq3_til(nf_list, n3lo_variation):
    """Initialize the CLq at N3LO tilde function from the file in External."""
    global CLq3_til
    CLq3_til = []
    append_empty(CLq3_til, nf_list)
    for nf in nf_list:
        CLq3_til_array = np.array(
            readt.readND(
                PATH_TO_GLOBAL
                + f"/External/CLq_3_til/CLq3til_nf{nf}_var{n3lo_variation}.txt"
            )
        )
        CLq3_til.append(interp2d(ZList[:-1], QList, CLq3_til_array.T, kind="quintic"))


def InitializeCg3_m(nf_list, n3lo_variation):
    """Initialize the Cg2 at N3LO massive function from the file in External."""
    global Cg3m
    Cg3m = []
    append_empty(Cg3m, nf_list)
    for nf in nf_list:
        Cg3m_array = np.array(
            readt.readND(
                PATH_TO_GLOBAL + f"/External/Cg_3_m/C2g_nf{nf}_var{n3lo_variation}.txt"
            )
        )
        Cg3m.append(interp2d(ZList[:-1], QList, Cg3m_array[:, :-1], kind="quintic"))


def InitializeCLg3_m(nf_list, n3lo_variation):
    """Initialize the CgL at N3LO massive function from the file in External."""
    global CLg3m
    CLg3m = []
    append_empty(CLg3m, nf_list)
    for nf in nf_list:
        CLg3m_array = np.array(
            readt.readND(
                PATH_TO_GLOBAL + f"/External/CLg_3_m/CLg_nf{nf}_var{n3lo_variation}.txt"
            )
        )
        CLg3m.append(interp2d(ZList, QList, CLg3m_array, kind="quintic"))


def InitializeCg3_til(nf_list, n3lo_variation):
    """Initialize the C2g at N3LO tilde function from the file in External."""
    global Cg3_til
    Cg3_til = []
    append_empty(Cg3_til, nf_list)
    for nf in nf_list:
        Cg3_til_array = np.array(
            readt.readND(
                PATH_TO_GLOBAL
                + f"/External/Cg_3_til/Cg3til_nf{nf}_var{n3lo_variation}.txt"
            )
        )
        Cg3_til.append(interp2d(ZList[:-1], QList, Cg3_til_array.T, kind="quintic"))


def InitializeCLg3_til(nf_list, n3lo_variation):
    """Initialize the CLg at N3LO tilde function from the file in External."""
    global CLg3_til
    CLg3_til = []
    append_empty(CLg3_til, nf_list)
    for nf in nf_list:
        CLg3_til_array = np.array(
            readt.readND(
                PATH_TO_GLOBAL
                + f"/External/CLg_3_til/CLg3til_nf{nf}_var{n3lo_variation}.txt"
            )
        )
        CLg3_til.append(interp2d(ZList[:-1], QList, CLg3_til_array.T, kind="quintic"))


def Initialize_all(nf=None, n3lo_variation=0):
    """Initialize all the needed global lists."""
    # TODO: allow also for Ftop
    nf_list = [nf] if nf is not None else [4, 5]
    console.log("[green]Loading precomputed grids ...")
    InitializeQX()
    InitializeMbg_3(nf_list)
    InitializeMbq_3(nf_list)
    InitializeCq3_m(nf_list, n3lo_variation)
    InitializeCLq3_m(nf_list, n3lo_variation)
    InitializeCg3_m(nf_list, n3lo_variation)
    InitializeCLg3_m(nf_list, n3lo_variation)
    InitializeCLq3_til(nf_list, n3lo_variation)
    InitializeCLg3_til(nf_list, n3lo_variation)
    InitializeCq3_til(nf_list, n3lo_variation)
    InitializeCg3_til(nf_list, n3lo_variation)


def InitializeHPL():
    global HPL_x_array
    global HPL_0011
    HPL_x_array = readt.read1D(PATH_TO_GLOBAL + "/External/HPL/HPL_x.txt")
    HPL_0011_array = readt.read1D(PATH_TO_GLOBAL + "/External/HPL/HPL_0011.txt")
    HPL_0011 = interp1d(HPL_x_array, HPL_0011_array)
    global HPL_00011
    HPL_00011_array = readt.read1D(PATH_TO_GLOBAL + "/External/HPL/HPL_00011.txt")
    HPL_00011 = interp1d(HPL_x_array, HPL_00011_array)
    global HPL_00101
    HPL_00101_array = readt.read1D(PATH_TO_GLOBAL + "/External/HPL/HPL_00101.txt")
    HPL_00101 = interp1d(HPL_x_array, HPL_00101_array)
    global HPL_00111
    HPL_00111_array = readt.read1D(PATH_TO_GLOBAL + "/External/HPL/HPL_00111.txt")
    HPL_00111 = interp1d(HPL_x_array, HPL_00111_array)
    global HPL_01011
    HPL_01011_array = readt.read1D(PATH_TO_GLOBAL + "/External/HPL/HPL_01011.txt")
    HPL_01011 = interp1d(HPL_x_array, HPL_01011_array)
    global HPL_0m1m1m1
    HPL_0m1m1m1_array = readt.read1D(PATH_TO_GLOBAL + "/External/HPL/HPL_0m1m1m1.txt")
    HPL_0m1m1m1 = interp1d(HPL_x_array, HPL_0m1m1m1_array)
    global HPL_0m101
    HPL_0m101_array = readt.read1D(PATH_TO_GLOBAL + "/External/HPL/HPL_0m101.txt")
    HPL_0m101 = interp1d(HPL_x_array, HPL_0m101_array)
    global HPL_00m1m1
    HPL_00m1m1_array = readt.read1D(PATH_TO_GLOBAL + "/External/HPL/HPL_00m1m1.txt")
    HPL_00m1m1 = interp1d(HPL_x_array, HPL_00m1m1_array)
    global HPL_00m11
    HPL_00m11_array = readt.read1D(PATH_TO_GLOBAL + "/External/HPL/HPL_00m11.txt")
    HPL_00m11 = interp1d(HPL_x_array, HPL_00m11_array)
    global HPL_001m1
    HPL_001m1_array = readt.read1D(PATH_TO_GLOBAL + "/External/HPL/HPL_001m1.txt")
    HPL_001m1 = interp1d(HPL_x_array, HPL_001m1_array)
    global HPL_0m10m1m1
    HPL_0m10m1m1_array = readt.read1D(PATH_TO_GLOBAL + "/External/HPL/HPL_0m10m1m1.txt")
    HPL_0m10m1m1 = interp1d(HPL_x_array, HPL_0m10m1m1_array)
    global HPL_00m1m1m1
    HPL_00m1m1m1_array = readt.read1D(PATH_TO_GLOBAL + "/External/HPL/HPL_00m1m1m1.txt")
    HPL_00m1m1m1 = interp1d(HPL_x_array, HPL_00m1m1m1_array)
    global HPL_00m10m1
    HPL_00m10m1_array = readt.read1D(PATH_TO_GLOBAL + "/External/HPL/HPL_00m10m1.txt")
    HPL_00m10m1 = interp1d(HPL_x_array, HPL_00m10m1_array)
    global HPL_00m101
    HPL_00m101_array = readt.read1D(PATH_TO_GLOBAL + "/External/HPL/HPL_00m101.txt")
    HPL_00m101 = interp1d(HPL_x_array, HPL_00m101_array)
    global HPL_000m1m1
    HPL_000m1m1_array = readt.read1D(PATH_TO_GLOBAL + "/External/HPL/HPL_000m1m1.txt")
    HPL_000m1m1 = interp1d(HPL_x_array, HPL_000m1m1_array)
    global HPL_000m11
    HPL_000m11_array = readt.read1D(PATH_TO_GLOBAL + "/External/HPL/HPL_000m11.txt")
    HPL_000m11 = interp1d(HPL_x_array, HPL_000m11_array)
    global HPL_0001m1
    HPL_0001m1_array = readt.read1D(PATH_TO_GLOBAL + "/External/HPL/HPL_0001m1.txt")
    HPL_0001m1 = interp1d(HPL_x_array, HPL_0001m1_array)
    global HPL_0010m1
    HPL_0010m1_array = readt.read1D(PATH_TO_GLOBAL + "/External/HPL/HPL_0010m1.txt")
    HPL_0010m1 = interp1d(HPL_x_array, HPL_0010m1_array)
    global HPL_0m1m11
    HPL_0m1m11_array = readt.read1D(PATH_TO_GLOBAL + "/External/HPL/HPL_0m1m11.txt")
    HPL_0m1m11 = interp1d(HPL_x_array, HPL_0m1m11_array)
    global HPL_0m11m1
    HPL_0m11m1_array = readt.read1D(PATH_TO_GLOBAL + "/External/HPL/HPL_0m11m1.txt")
    HPL_0m11m1 = interp1d(HPL_x_array, HPL_0m11m1_array)
    global HPL_01m1m1
    HPL_01m1m1_array = readt.read1D(PATH_TO_GLOBAL + "/External/HPL/HPL_01m1m1.txt")
    HPL_01m1m1 = interp1d(HPL_x_array, HPL_01m1m1_array)
    global HPL_0m111
    HPL_0m111_array = readt.read1D(PATH_TO_GLOBAL + "/External/HPL/HPL_0m111.txt")
    HPL_0m111 = interp1d(HPL_x_array, HPL_0m111_array)
    global HPL_01m11
    HPL_01m11_array = readt.read1D(PATH_TO_GLOBAL + "/External/HPL/HPL_01m11.txt")
    HPL_01m11 = interp1d(HPL_x_array, HPL_01m11_array)
    global HPL_011m1
    HPL_011m1_array = readt.read1D(PATH_TO_GLOBAL + "/External/HPL/HPL_011m1.txt")
    HPL_011m1 = interp1d(HPL_x_array, HPL_011m1_array)
    global HPL_0m1011
    HPL_0m1011_array = readt.read1D(PATH_TO_GLOBAL + "/External/HPL/HPL_0m1011.txt")
    HPL_0m1011 = interp1d(HPL_x_array, HPL_0m1011_array)
    global HPL_0m1m101
    HPL_0m1m101_array = readt.read1D(PATH_TO_GLOBAL + "/External/HPL/HPL_0m1m101.txt")
    HPL_0m1m101 = interp1d(HPL_x_array, HPL_0m1m101_array)
    global HPL_0m1m11m1
    HPL_0m1m11m1_array = readt.read1D(PATH_TO_GLOBAL + "/External/HPL/HPL_0m1m11m1.txt")
    HPL_0m1m11m1 = interp1d(HPL_x_array, HPL_0m1m11m1_array)
    global HPL_0m10m11
    HPL_0m10m11_array = readt.read1D(PATH_TO_GLOBAL + "/External/HPL/HPL_0m10m11.txt")
    HPL_0m10m11 = interp1d(HPL_x_array, HPL_0m10m11_array)
    global HPL_0m101m1
    HPL_0m101m1_array = readt.read1D(PATH_TO_GLOBAL + "/External/HPL/HPL_0m101m1.txt")
    HPL_0m101m1 = interp1d(HPL_x_array, HPL_0m101m1_array)
    global HPL_0m11m1m1
    HPL_0m11m1m1_array = readt.read1D(PATH_TO_GLOBAL + "/External/HPL/HPL_0m11m1m1.txt")
    HPL_0m11m1m1 = interp1d(HPL_x_array, HPL_0m11m1m1_array)
    global HPL_00m1m11
    HPL_00m1m11_array = readt.read1D(PATH_TO_GLOBAL + "/External/HPL/HPL_00m1m11.txt")
    HPL_00m1m11 = interp1d(HPL_x_array, HPL_00m1m11_array)
    global HPL_00m11m1
    HPL_00m11m1_array = readt.read1D(PATH_TO_GLOBAL + "/External/HPL/HPL_00m11m1.txt")
    HPL_00m11m1 = interp1d(HPL_x_array, HPL_00m11m1_array)
    global HPL_00m111
    HPL_00m111_array = readt.read1D(PATH_TO_GLOBAL + "/External/HPL/HPL_00m111.txt")
    HPL_00m111 = interp1d(HPL_x_array, HPL_00m111_array)
    global HPL_001m1m1
    HPL_001m1m1_array = readt.read1D(PATH_TO_GLOBAL + "/External/HPL/HPL_001m1m1.txt")
    HPL_001m1m1 = interp1d(HPL_x_array, HPL_001m1m1_array)
    global HPL_001m11
    HPL_001m11_array = readt.read1D(PATH_TO_GLOBAL + "/External/HPL/HPL_001m11.txt")
    HPL_001m11 = interp1d(HPL_x_array, HPL_001m11_array)
    global HPL_0011m1
    HPL_0011m1_array = readt.read1D(PATH_TO_GLOBAL + "/External/HPL/HPL_0011m1.txt")
    HPL_0011m1 = interp1d(HPL_x_array, HPL_0011m1_array)
    global HPL_01m1m1m1
    HPL_01m1m1m1_array = readt.read1D(PATH_TO_GLOBAL + "/External/HPL/HPL_01m1m1m1.txt")
    HPL_01m1m1m1 = interp1d(HPL_x_array, HPL_01m1m1m1_array)
    global HPL_0m1m1m1m1
    HPL_0m1m1m1m1_array = readt.read1D(
        PATH_TO_GLOBAL + "/External/HPL/HPL_0m1m1m1m1.txt"
    )
    HPL_0m1m1m1m1 = interp1d(HPL_x_array, HPL_0m1m1m1m1_array)
    global HPL_0m1m1m11
    HPL_0m1m1m11_array = readt.read1D(PATH_TO_GLOBAL + "/External/HPL/HPL_0m1m1m11.txt")
    HPL_0m1m1m11 = interp1d(HPL_x_array, HPL_0m1m1m11_array)



import numpy as np
from eko.mellin import Path
from ekore.harmonics import compute_cache
from scipy import integrate


def check_path(r, s):
    if r + s < 1.0:
        raise ValueError(
            "r + s must be greater than 1 (i.e. than the real part of the rightmost pole)"
        )


def quad_ker_talbot(u, func, x, nf, L):
    is_singlet = True
    path = Path(u, np.log(x), is_singlet)
    integrand = path.prefactor * x ** (-path.n) * path.jac
    sx = compute_cache(path.n, 5, is_singlet=is_singlet)
    sx = [np.array(s) for s in sx]
    try:
        gamma = func(path.n, sx, nf, L)
    except TypeError:
        gamma = func(path.n, sx, L)
    return np.real(gamma * integrand)


def inverse_mellin_talbot(func, x, nf, L):
    mellin_cut = 5e-2
    return integrate.quad(
        lambda u: quad_ker_talbot(u, func, x, nf, L),
        0.5,
        1.0 - mellin_cut,
        epsabs=1e-12,
        epsrel=1e-10,
        limit=200,
        full_output=1,
    )[0]


def path_linear(t, r, s):
    return complex(s - r * t, t)


def jac_linear(r):
    return complex(-r, 1)


def quad_ker_linear(t, func, x, nf, r, s, L):
    n = path_linear(t, r, s)
    jac = jac_linear(r)
    sx = compute_cache(n, 5, is_singlet=True)
    sx = [np.array(s) for s in sx]
    gamma = func(n, sx, nf, L)
    return np.imag(x ** (-n) * gamma * jac) / np.pi


def inverse_mellin_linear(func, x, nf, r, s, L):
    r = 1.0 if r is None else r
    s = 1.5 if s is None else s
    check_path(0, s)
    return integrate.quad(
        lambda u: quad_ker_linear(u, func, x, nf, r, s, L),
        0,
        np.inf,
        epsabs=1e-12,
        epsrel=1e-10,
        limit=200,
        full_output=1,
    )[0]


def inverse_mellin(func, x, nf, r, s, path, L):
    if path == "talbot":
        return inverse_mellin_talbot(func, x, nf, L)
    elif path == "linear":
        return inverse_mellin_linear(func, x, nf, r, s, L)
    else:
        raise NotImplementedError(
            "Selected path is not implemented: choose either 'talbot' or 'linear'"
        )



"""Tilde coefficents functions for the matched scheme for the light components and asymptotics."""
import numpy as np

from .MasslessCoeffFunc import (
    Cb_1_loc,
    Cb_1_reg,
    Cb_1_sing,
    Cb_2_loc,
    Cb_2_reg,
    Cb_2_sing,
    Cb_3_loc,
    Cb_3_reg,
    Cb_3_sing,
    
    Cg_2_reg,
    Cq_2_reg,
    Cg_3_reg,
    Cg_3_loc,
    Cq_3_loc,
    Cq_3_reg,

    CLb_1_reg,
    CLb_2_loc,
    CLb_2_reg,
    CLb_3_loc,
    CLb_3_reg,

    CLq_2_reg,
    CLg_2_reg,
    CLq_3_reg,
    CLg_3_reg,
)
from .MatchingFunc import (
    P1,
    Mqq_2_reg,
    Mqq_2_loc,
    Mqq_2_sing,
)


from yadism.coefficient_functions.fonll import raw_nc


############ NNLO Massive Non Singlet Asymptotics ########

# F2
def Cb_2_asy_reg(z, Q, p, _nf):
    e_q_light = p[-1] ** 2
    L = np.log((p[1] ** 2) / (p[0] ** 2))
    return e_q_light * (
        raw_nc.c2ns2am0_aq2(z) * L**2
        + raw_nc.c2ns2am0_aq(z) * L
        + raw_nc.c2ns2am0_a0(z)
    )

def Cb_2_asy_loc(z, Q, p, _nf):
    e_q_light = p[-1] ** 2
    L = np.log((p[1] ** 2) / (p[0] ** 2))
    return e_q_light * (
        raw_nc.c2ns2cm0_aq2(z) * L**2
        + raw_nc.c2ns2cm0_aq(z) * L
        + raw_nc.c2ns2cm0_a0(z)
    )

def Cb_2_asy_sing(z, Q, p, _nf):
    e_q_light = p[-1] ** 2
    L = np.log((p[1] ** 2) / (p[0] ** 2))
    return e_q_light * (
        raw_nc.c2ns2bm0_aq2(z) * L**2
        + raw_nc.c2ns2bm0_aq(z) * L
        + raw_nc.c2ns2bm0_a0(z)
    )

# FL
def CLb_2_asy_reg(z, Q, p, _nf):
    e_q_light = p[-1] ** 2
    L = np.log((p[1] ** 2) / (p[0] ** 2))
    return e_q_light * (
        + raw_nc.clns2am0_aq(z) * L
        + raw_nc.clns2am0_a0(z)
    )

############ NNLO and N3LO Tilde coefficients ############
# NOTE: at LO, NLO all these functions 
# should trivially reduce to the massless ones.

##### F2 NNLO
# NOTE: due to some cancellations see eq 96 and 97 of
# https://arxiv.org/pdf/1001.2312.pdf, in NS
# we just need to evaluate the massless at nf = nl
def Cb_2_til_reg(z, Q, p, nl):
    e_q_light = p[-1] ** 2
    return Cb_2_reg(z, Q, p, nl) -(e_q_light * Mqq_2_reg(z, p, nl) + P1(p, nl) * Cb_1_reg(z, Q, p, nl))


def Cb_2_til_loc(z, Q, p, nl):
    e_q_light = p[-1] ** 2
    return Cb_2_loc(z, Q, p, nl) -(e_q_light * Mqq_2_loc(z, p, nl) + P1(p, nl) * Cb_1_loc(z, Q, p, nl))


def Cb_2_til_sing(z, Q, p, nl):
    e_q_light = p[-1] ** 2
    return Cb_2_sing(z, Q, p, nl) -(e_q_light * Mqq_2_sing(z, p, nl) + P1(p, nl) * Cb_1_sing(z, Q, p, nl))

def Cg_2_til_reg(z, Q, p, _nl):
    return Cg_2_reg(z, Q, p, _nl + 1)


def Cq_2_til_reg(z, Q, p, _nl):
    return Cq_2_reg(z, Q, p, _nl + 1)

#### F2 N3LO
# TODO: construct the full N3LO tilde once massive 
# coefficient functions to light will be available
# for the moment everthin reduces to the pure massless in nf = nl+1
def Cb_3_til_reg(z, Q, p, nl):
    return Cb_3_reg(z, Q, p, nl + 1)

def Cb_3_til_loc(z, Q, p, nl):
    return Cb_3_loc(z, Q, p, nl + 1)

def Cb_3_til_sing(z, Q, p, nl):
    return Cb_3_sing(z, Q, p, nl + 1)

def Cg_3_til_reg(z, Q, p, nl):
    return Cg_3_reg(z, Q, p, nl + 1)

def Cg_3_til_loc(z, Q, p, nl):
    return Cg_3_loc(z, Q, p, nl + 1)

def Cq_3_til_reg(z, Q, p, nl):
    return Cq_3_reg(z, Q, p, nl + 1)

def Cq_3_til_loc(z, Q, p, nl):
    return Cq_3_loc(z, Q, p, nl + 1)


#### FL NNLO
def CLb_2_til_reg(z, Q, p, nl):
    return CLb_2_reg(z, Q, p, nl) - P1(p, nl) * CLb_1_reg(z, Q, p, nl)

def CLb_2_til_loc(z, Q, p, _nl):
    return CLb_2_loc(z, Q, p, _nl)


def CLg_2_til_reg(z, Q, p, _nl):
    return CLg_2_reg(z, Q, p, _nl + 1)


def CLq_2_til_reg(z, Q, p, _nl):
    return CLq_2_reg(z, Q, p, _nl + 1)

#### FL N3LO
def CLb_3_til_reg(z, Q, p, nl):
    return CLb_3_reg(z, Q, p, nl + 1)

def CLb_3_til_loc(z, Q, p, nl):
    return CLb_3_loc(z, Q, p, nl + 1)


def CLg_3_til_reg(z, Q, p, nl):
    return CLg_3_reg(z, Q, p, nl + 1)


def CLq_3_til_reg(z, Q, p, nl):
    return CLq_3_reg(z, Q, p, nl + 1)



"""Tools related to the configuration file handling."""
import copy
import pathlib

import tomli

name = "dis_tp.toml"
"Name of the config file (wherever it is placed)"

# better to declare immediately the correct type
configs = {}
"Holds loaded configurations"


def defaults(base_configs):
    """Provide additional defaults.

    Parameters
    ----------
    base_config : dict
        user provided configuration

    Returns
    -------
    configs_ : dict
        enhanced configuration

    Note
    ----
    The general rule is to never replace user provided input.
    """
    configs_ = copy.deepcopy(base_configs)
    enhance_paths(configs_)

    return configs_


def enhance_paths(configs_):
    """Check required path and enhance them with root path.

    The changes are done inplace.

    Parameters
    ----------
    configs_ : dict
        configuration
    """
    # required keys without default
    for key in ["theory_cards", "operator_cards", "results"]:
        if key not in configs_["paths"]:
            raise ValueError(f"Configuration is missing a 'paths.{key}' key")
        if pathlib.Path(configs_["paths"][key]).anchor == "":
            configs_["paths"][key] = configs_["paths"]["root"] / configs_["paths"][key]
        else:
            configs_["paths"][key] = pathlib.Path(configs_["paths"][key])


def detect(path=None):
    """Autodetect configuration file path.

    Parameters
    ----------
    path : str or os.PathLike
        user provided guess

    Returns
    -------
    pathlib.Path :
        file path
    """
    # If a path is provided then we only look for the dis_tp.toml file there.
    if path is not None:
        path = pathlib.Path(path)
        configs_file = path / name if path.is_dir() else path
        if configs_file.is_file():
            return configs_file
        else:
            raise ValueError(
                "Provided path is not pointing to (or does not contain) the dis_tp.toml file"
            )

    # If no path is provided we need to look after the file.
    # We want to check cwd and all its parent folders (without their subfolders
    # of course) up to root.
    cwd_path = pathlib.Path.cwd().absolute()
    paths = [cwd_path] + list(cwd_path.parents)

    for p in paths:
        configs_file = p / name

        if configs_file.is_file():
            return configs_file

    raise FileNotFoundError("No configurations file detected.")


def load(path=None):
    """Load config file.

    Parameters
    ----------
    path : str or os.PathLike
        file path

    Returns
    -------
    loaded : dict
        configuration dictionary
    """
    path = detect(path)

    with open(path, "rb") as fd:
        loaded = tomli.load(fd)

    if "paths" not in loaded:
        loaded["paths"] = {}
    if "root" not in loaded["paths"]:
        loaded["paths"]["root"] = pathlib.Path(path).parent

    return loaded



import pathlib

import click
import numpy as np
import yaml

from .. import configs, io, k_factors, parameters, plot, runner
from .base import command
from .grids import n_cores


def provide_default_kinematics(obs):
    """Provide default kinematics for an observable."""
    h_id = 4 if "charm" in obs else 5
    mass = parameters.masses(h_id)
    Qlogmin = np.log10(1.0)
    Qlogmax = np.log10(150.0)
    Qlog = np.linspace(Qlogmin, Qlogmax, 200)
    Qcommon = pow(10, Qlog)
    eps = 0.5
    thre = [mass * ratio for ratio in [0.5, 1.0, 2.0]]
    Qsing = [np.linspace(thr - eps, thr + eps, 5) for thr in thre]
    Q = np.sort(np.concatenate((Qcommon, Qsing[0], Qsing[1], Qsing[2]))).tolist()
    X = [0.1, 0.01, 0.001, 0.0001]
    kinematics = []
    for x in X:
        for q in Q:
            dictx = {}
            dictx["x"] = x
            dictx["q"] = q
            kinematics.append(dictx)
    return kinematics


t_card = click.argument("t_card", type=str)
o_card = click.argument("o_card", type=str)
obs = click.argument("obs", type=str)
restype = click.argument("restype", type=str)
pdf = click.argument("pdf", type=str)
scalevar = click.argument("scalevar", type=bool)
plot_dir = click.argument("plot_dir", type=str)
order = click.argument("order", type=str)
h_id = click.argument("h_id", type=str)
cfg = click.option(
    "-c",
    "--configs",
    "cfg",
    default=None,
    type=click.Path(resolve_path=True, path_type=pathlib.Path),
    help="Explicitly specify config file (it has to be a valid TOML file).",
)


@command.command("compute")
@o_card
@t_card
@n_cores
@cfg
def generate_matching_grids(o_card: str, t_card: str, n_cores: int, cfg: pathlib.Path):
    """
    Run a computation.

    USAGE dis_tp compute <o_card> <t_card>
    """

    obj = runner.Runner(o_card, t_card, cfg)
    obj.compute(n_cores)
    obj.save_results()


@command.command("add_observable")
@o_card
@obs
@restype
@pdf
@scalevar
@cfg
def add_obs_opcard(
    o_card: str, obs: str, restype: str, pdf: str, scalevar: bool, cfg: pathlib.Path
):
    """
    Add an observable to the operator card with default kinematics.

    USAGE dis_tp add_observable <o_card> <obs> <restype> <pdf> <scalevar>
    """

    cfg = configs.load(cfg)
    cfg = configs.defaults(cfg)
    ocard_path = cfg["paths"]["operator_cards"] / (o_card + ".yaml")
    old_ocard = {}
    if ocard_path.is_file():
        with open(
            cfg["paths"]["operator_cards"] / (o_card + ".yaml"), encoding="utf-8"
        ) as f:
            old_ocard = yaml.safe_load(f)
    kinematics = provide_default_kinematics(obs)
    to_update = {
        "obs": {
            obs: dict(
                PDF=pdf, restype=restype, scalevar=scalevar, kinematics=kinematics
            )
        }
    }
    old_ocard["obs"].update(to_update["obs"])
    with open(
        cfg["paths"]["operator_cards"] / (o_card + ".yaml"), "w", encoding="utf-8"
    ) as f:
        yaml.safe_dump(old_ocard, f)


@command.command("plot")
@plot_dir
@obs
@order
@h_id
@cfg
def plot_observable(plot_dir: str, obs: str, order: str, h_id: str, cfg: pathlib.Path):
    """
    Plot (and save the results in plot_dir) an observable with all the method overimposed for a certain order
    and heavy quark id.

    USAGE dis_tp plot <plot_dir> <obs> <order> <h_id>
    """
    cfg = configs.load(cfg)
    cfg = configs.defaults(cfg)
    plot_dir_path = cfg["paths"]["root"] / plot_dir
    plotclass = plot.Plot(cfg, plot_dir_path)
    orders = {"NLO": 1, "NNLO": 2, "N3LO": 3}
    plotclass.plot_single_obs(obs, str(orders[order]), h_id)
    # plotclass.plot_fonll_noerr(obs, str(orders[order]), h_id)
    # plotclass.plot_fonll_order_comparison(obs, h_id)


@command.command("k-factors")
@o_card
@pdf
@n_cores
@click.argument(
    "author",
    type=str,
)
@click.argument(
    "t_cards",
    nargs=-1,
    type=str,
)
@click.option(
    "-yad",
    "--use_yadism",
    is_flag=True,
    type=bool,
    default=False,
    required=False,
    help="If True compute the k-factor w.r.t. Yadism",
)
@click.option(
    "--fonll_incomplete",
    type=bool,
    default=False,
    is_flag=True,
    required=False,
    help="""Use DIS TP FONLL_incomplete in the denominator.""",
)
@click.option(
    "-th",
    "--th_description",
    type=str,
    default="NNPDF4.0 pch with alphas(MZ)=0.118",
    required=False,
    help="TheoryInput to be stored in the CF file",
)
@cfg
def generate_kfactors(
    o_card: str,
    t_cards: str,
    pdf: str,
    author: str,
    n_cores: int,
    use_yadism: bool,
    fonll_incomplete: bool,
    th_description: str,
    cfg: pathlib.Path,
):
    """Generate k-factors.

    USAGE: dis_tp k-factors HERA_NC_318GEV_EAVG_SIGMARED_CHARM NNPDF40_nnlo_pch_as_01180 "Your Name" 439 518 [--fonll_incomplete -n 4 -yad -th "Theory Input"]
    """

    result_logs = []
    for t_card in t_cards:
        obj = k_factors.KfactorRunner(
            t_card, o_card, pdf, use_yadism, fonll_incomplete, cfg
        )
        obj.compute(n_cores)
        if len(t_cards) == 1:
            # No variations
            obj.save_results(author, th_input=th_description)
            exit(0)
        else:
            # Append all the variations
            result_logs.append(obj._results)
    # save the k factors
    obj._results = obj.build_kfactor_with_error(result_logs)
    obj.save_results(author, th_input=th_description)



# This contains my implementation in python of the harmonic polylogs. Some of them are written in terms of polylogs, others are taken from mathematica grids.
import numpy as np
from eko.harmonics.constants import zeta3, zeta5
from mpmath import *

from . import Initialize as Ini

mp.dps = 15
mp.pretty = True


def H_0(x):
    return np.log(x)


def H_1(x):
    return -np.log(1 - x)


def H_10(x):
    return -np.log(1 - x) * np.log(x) - polylog(2, x)


def H_m10(x):
    return np.log(1 + x) * np.log(x) + polylog(2, -x)


def H_00(x):
    return (np.log(x) ** 2) / 2.0


def H_01(x):
    return polylog(2, x)


def H_001(x):
    return polylog(3, x)


def H_011(x):
    return (
        (1.0 / 2.0) * (np.log(x)) * (np.log(1 - x) ** 2)
        + (np.log(1 - x)) * polylog(2, 1 - x)
        - polylog(3, 1 - x)
        + zeta3
    )


def H_0001(x):
    return polylog(4, x)


def H_0011(x):  ##
    return Ini.HPL_0011(x)


def H_0111(x):
    return (
        (np.pi**4) / (90.0)
        - (1.0 / 6.0) * (np.log(1 - x) ** 3) * np.log(x)
        - (1.0 / 2.0) * (np.log(1 - x) ** 2) * polylog[2, 1 - x]
        + np.log(1 - x) * polylog[3, 1 - x]
        - polylog[4, 1 - x]
    )


def H_00001(x):
    return polylog(5, x)


def H_00011(x):  ##
    return Ini.HPL_00011(x)


def H_00101(x):  ##
    return Ini.HPL_00101(x)


def H_00111(x):  ##
    return Ini.HPL_00111(x)


def H_01011(x):  ##
    return Ini.HPL_01011(x)


def H_01111(x):
    return (
        (1.0 / 24.0) * (np.log(1 - x) ** 4) * np.log(x)
        + (1.0 / 6.0) * (np.log(1 - x) ** 3) * polylog(2, 1 - x)
        - (1.0 / 2.0) * (np.log(1 - x) ** 2) * polylog(3, 1 - x)
        + np.log(1 - x) * polylog(4, 1 - x)
        - polylog(5, 1 - x)
        + zeta5
    )


def H_0111(x):
    return (
        (np.pi**4) / (90.0)
        - (1.0 / 6.0) * (np.log(1 - x) ** 3) * np.log(x)
        - (1.0 / 2.0) * (np.log(1 - x) ** 2) * polylog(2, 1 - x)
        + np.log(1 - x) * polylog(3, 1 - x)
        - polylog(4, 1 - x)
    )


def H_0m1(x):
    return -polylog(2, -x)


def H_m1(x):
    return np.log(1 + x)


def H_00m1(x):
    return -polylog(3, -x)


def H_0m1m1(x):
    return S12(-x)


def H_0m11(x):
    return (
        ((np.pi**2) / (12.0) - (np.log(2) ** 2) / (2.0)) * np.log(1 - x)
        + (1.0 / 2.0) * np.log(2) * (np.log(1 - x) ** 2)
        - (1.0 / 6.0) * (np.log(1 - x) ** 3)
        - (1.0 / 2.0) * (np.log(1 - x) ** 2) * np.log(x)
        - np.log(1 - x) * polylog(2, 1 - x)
        + np.log(1 - x) * polylog(2, -x)
        - np.log(1 - x) * polylog(2, x)
        - polylog(3, (1 - x) / 2.0)
        + polylog(3, 1 - x)
        - polylog(3, -x)
        + polylog(3, x)
        + polylog(3, -((2 * x) / (1 - x)))
        - zeta3
        + (1.0 / 24.0)
        * (-2 * (np.pi**2) * np.log(2) + 4 * (np.log(2) ** 3) + 21 * zeta3)
    )


def H_01m1(x):
    return (
        ((np.pi**2) / (12.0) - (np.log(2) ** 2) / (2.0)) * np.log(1 + x)
        + (1.0 / 2.0) * np.log(2) * (np.log(1 + x) ** 2)
        + np.log(1 + x) * polylog(2, x)
        - polylog(3, x)
        - polylog(3, (x) / (1 + x))
        + polylog(3, (2 * x) / (1 + x))
        - polylog(3, (1 + x) / 2)
        + (1.0 / 24.0)
        * (-2 * (np.pi**2) * np.log(2) + 4 * (np.log(2) ** 3) + 21 * zeta3)
    )


def H_0m1m1m1(x):  ## ok but there was imaginary part---> ask
    return Ini.HPL_0m1m1m1(x)


def H_0m101(x):  ##
    return Ini.HPL_0m101(x)


def H_00m1m1(x):  ##
    return Ini.HPL_00m1m1(x)


def H_00m11(x):  ##
    return Ini.HPL_00m11(x)


def H_000m1(x):
    return -polylog(4, -x)


def H_001m1(x):  ##
    return Ini.HPL_001m1(x)


def H_0m10m1m1(x):  ##
    return Ini.HPL_0m10m1m1(x)


def H_00m1m1m1(x):  ##
    return Ini.HPL_00m1m1m1(x)


def H_00m10m1(x):  ##
    return Ini.HPL_00m10m1(x)


def H_00m101(x):  ##
    return Ini.HPL_00m101(x)


def H_000m1m1(x):  ##
    return Ini.HPL_000m1m1(x)


def H_000m11(x):  ##
    return Ini.HPL_000m11(x)


def H_0000m1(x):
    return -polylog(5, -x)


def H_0001m1(x):  ##
    return Ini.HPL_0001m1(x)


def H_0010m1(x):  ##
    return Ini.HPL_0010m1(x)


def H_0m1m11(x):  ##
    return Ini.HPL_0m1m11(x)


def H_0m11m1(x):  ##
    return Ini.HPL_0m11m1(x)


def H_01m1m1(x):  ##
    return Ini.HPL_01m1m1(x)


def H_0m111(x):  ##
    return Ini.HPL_0m111(x)


def H_01m11(x):  ##
    return Ini.HPL_01m11(x)


def H_011m1(x):  ##
    return Ini.HPL_011m1(x)


def H_0m1011(x):  ##
    return Ini.HPL_0m1011(x)


def H_0m1m101(x):  #
    return Ini.HPL_0m1m101(x)


def H_0m1m1m1m1(x):  #
    return Ini.HPL_0m1m1m1m1(x)


def H_0m1m1m11(x):  ##
    return Ini.HPL_0m1m1m11(x)


def H_0m1m11m1(x):  ##
    return Ini.HPL_0m1m11m1(x)


def H_0m10m11(x):  ##
    return Ini.HPL_0m10m11(x)


def H_0m101m1(x):  ##
    return Ini.HPL_0m101m1(x)


def H_0m11m1m1(x):  ##
    return Ini.HPL_0m11m1m1(x)


def H_00m1m11(x):  ##
    return Ini.HPL_00m1m11(x)


def H_00m11m1(x):  ##
    return Ini.HPL_00m11m1(x)


def H_00m111(x):  ##
    return Ini.HPL_00m111(x)


def H_001m1m1(x):  ##
    return Ini.HPL_001m1m1(x)


def H_001m11(x):  ##
    return Ini.HPL_001m11(x)


def H_0011m1(x):  ##
    return Ini.HPL_0011m1(x)


def H_01m1m1m1(x):  ##
    return Ini.HPL_01m1m1m1(x)


# Generalized Nielsen
def S12(x):
    ZETA3 = zeta3
    if x > 1:
        return (
            -polylog(3, (1.0 - x))
            + ZETA3
            + np.log(x - 1.0) * polylog(2, (1.0 - x))
            + 0.5 * np.log(x) * (np.log(x - 1.0) * np.log(x - 1.0) - np.pi * np.pi)
        )

    elif x == 1:
        return ZETA3

    elif (0.0 < x) and (x < 1):
        return (
            -polylog(3, (1.0 - x))
            + ZETA3
            + np.log(1.0 - x) * polylog(2, (1.0 - x))
            + 0.5 * np.log(x) * np.log(1.0 - x) * np.log(1.0 - x)
        )

    elif x == 0:
        return 0.0

    elif x < 0.0:
        c = 1.0 / (1.0 - x)
        return (
            -polylog(3, c)
            + ZETA3
            + np.log(c) * polylog(2, c)
            + 0.5 * np.log(c) * np.log(c) * np.log(1.0 - c)
            - (1.0 / 6.0) * np.log(c) * np.log(c) * np.log(c)
        )

    else:
        return 0.0



import numpy as np

from eko.thresholds import ThresholdsAtlas
import yadism.coefficient_functions.coupling_constants as coupl
from yadism.coefficient_functions.light import n3lo

pids = {"g": 21, "c": 4, "b": 5, "t": 6}


def number_active_flavors(Q):
    return _thr_atlas.nf(Q**2)


def number_light_flavors(Q):
    """This should match the FONLL prescription."""
    nf = _thr_atlas.nf(Q**2)
    if nf > 3:
        return nf - 1
    return nf


def charges(h_id):
    ch = {
        1: -1.0 / 3.0,
        2: 2.0 / 3.0,
        3: -1.0 / 3.0,
        4: 2.0 / 3.0,
        5: -1.0 / 3.0,
        6: 2.0 / 3.0,
    }
    return np.sign(h_id) * ch[h_id]


def default_masses(h_id):
    m = {4: 1.51, 5: 4.92, 6: 172.5}
    return m[h_id]


def initialize_theory(use_grids, masses=None, strong_coupling=None, thr_atlas=None):
    if not use_grids and masses is None:
        raise ValueError(
            f"Need to specify heavy particle masses when grids are not used."
        )
    if use_grids and masses is not None:
        for i, mass in enumerate(masses):
            if not np.isclose(mass, default_masses(i + 4)):
                raise ValueError(
                    f"Grids are only available for the default mass {default_masses(i+4)}."
                )
    global grids
    global _masses
    grids = use_grids
    if masses is None:
        _masses = [default_masses(4), default_masses(5), default_masses(6)]
    else:
        _masses = masses

    if strong_coupling is not None:
        global _alpha_s
        _alpha_s = strong_coupling.a_s

    global _thr_atlas
    # enforce some defaults
    if thr_atlas is None:
        _thr_atlas = ThresholdsAtlas(
            masses=np.array(_masses) ** 2,
            thresholds_ratios=[1,1,1]
        )
    else:
        _thr_atlas = thr_atlas

def masses(h_id):
    return _masses[h_id - 4]


def alpha_s(mur2, q2):
    return _alpha_s(mur2, q2)

# some default values for EM (therory is ignored)
_th_d = dict(
        SIN2TW=0.23126,
        MZ=91.1876,
        CKM="0.97428 0.22530 0.003470 0.22520 0.97345 0.041000 0.00862 0.04030 0.999152",
)
obs_d = dict(
        projectilePID=11,
        PolarizationDIS=0.0,
        prDIS="EM",
        PropagatorCorrection=0,
        NCPositivityCharge=None,
)
coupl_const = coupl.CouplingConstants.from_dict(_th_d, obs_d)
def n3lo_color_factors(partonic_channel, nf, skip_heavylight):
    """Compute N3LO color facotrs. nf is the number of total active flavors"""
    return n3lo.common.nc_color_factor(coupl_const, nf, partonic_channel, skip_heavylight)


import pathlib

import matplotlib.gridspec as gridspec
import matplotlib.pyplot as plt
import numpy as np
import yaml
from matplotlib import pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages

from . import io, parameters

orderstrings = {"1": "nlo", "2": "nnlo", "3": "nnlo"}
heavy_dict = {"4": "charm", "5": "bottom"}


class Plot:
    """Class for handling plots"""

    def __init__(self, configs: dict, plot_dir: pathlib.Path):
        self.result_path = configs["paths"]["results"]
        self.plot_dir = plot_dir

    def get_restypes(self, order, h_id, restype):
        restypes = {
            "FO": {
                "color": "violet",
                "pdf": "NNPDF40_"
                + orderstrings[order]
                + "_as_01180_nf_"
                + str(int(h_id) - 1),
            },
            "M": {
                "color": "blue",
                "pdf": [
                    "NNPDF_" + h_id + "F_" + orderstrings[order] + "_mub=05mb",
                    "NNPDF_" + h_id + "F_" + orderstrings[order],
                    "NNPDF_" + h_id + "F_" + orderstrings[order] + "_mub=2mb",
                ],
            },
            "R": {
                "color": "green",
                "pdf": [
                    "NNPDF_" + h_id + "F_" + orderstrings[order] + "_mub=05mb",
                    "NNPDF_" + h_id + "F_" + orderstrings[order],
                    "NNPDF_" + h_id + "F_" + orderstrings[order] + "_mub=2mb",
                ],
            },
        }
        return restypes[restype]

    def x_q_grids(self, obs, order, h_id):
        _, x_grid, q_grid = self.get_FO_result(obs, order, h_id)
        return (x_grid, q_grid)

    def get_FO_result(self, obs, order, h_id):
        mu_list = ["0.5", "1.0", "2.0"]
        filenames_FO = [
            obs
            + "_"
            + "FO"
            + "_"
            + order
            + "_"
            + h_id
            + "_"
            + heavy_dict[h_id]
            + "_thr="
            + thr
            + "_"
            + self.get_restypes(order, h_id, "FO")["pdf"]
            + "_0"
            for thr in mu_list
        ]
        results_FO = {}
        for filepath, mu in zip(filenames_FO, mu_list):
            with open(self.result_path / (filepath + ".yaml"), encoding="utf-8") as f:
                results_FO[mu] = yaml.safe_load(f)
        x_grid = results_FO[mu_list[1]]["x_grid"]
        q_grid = results_FO[mu_list[1]]["q_grid"]
        ordered_results_FO = {}
        for mu in mu_list:
            tmp_list = []
            for x, q, res in zip(x_grid, q_grid, results_FO[mu]["obs"][0]):
                tmp_list.append(dict(x=x, q=q, res=res))
            ordered_results_FO[mu] = tmp_list
        return (ordered_results_FO, x_grid, q_grid)

    def get_FO_n3lo_var_results(self, obs, h_id):
        n3lo_cf_variations = ["-1", "1"]
        filenames_FO = [
            obs
            + "_"
            + "FO"
            + "_"
            + "3"
            + "_"
            + h_id
            + "_"
            + heavy_dict[h_id]
            + "_thr="
            + "1.0"
            + "_"
            + self.get_restypes("3", h_id, "FO")["pdf"]
            + "_"
            + var
            for var in n3lo_cf_variations
        ]
        results_FO = {}
        for filepath, var in zip(filenames_FO, n3lo_cf_variations):
            with open(self.result_path / (filepath + ".yaml"), encoding="utf-8") as f:
                results_FO[var] = yaml.safe_load(f)
        x_grid = results_FO[n3lo_cf_variations[0]]["x_grid"]
        q_grid = results_FO[n3lo_cf_variations[0]]["q_grid"]
        ordered_results_FO = {}
        for var in n3lo_cf_variations:
            tmp_list = []
            for x, q, res in zip(x_grid, q_grid, results_FO[var]["obs"][0]):
                tmp_list.append(dict(x=x, q=q, res=res))
            ordered_results_FO[var] = tmp_list
        return ordered_results_FO

    def get_R_results(self, obs, order, h_id):
        mu_list = ["0.5", "1.0", "2.0"]
        filenames_R = [
            obs
            + "_"
            + "R"
            + "_"
            + order
            + "_"
            + h_id
            + "_"
            + heavy_dict[h_id]
            + "_thr="
            + mu
            + "_"
            + pdf
            + "_0"
            for pdf, mu in zip(self.get_restypes(order, h_id, "R")["pdf"], mu_list)
        ]
        results_R = {}
        for filepath in filenames_R:
            with open(self.result_path / (filepath + ".yaml"), encoding="utf-8") as f:
                string = filepath.split(orderstrings[order])[-1].split("_0")[0]
                results_R[string] = yaml.safe_load(f)
        x_grid, q_grid = self.x_q_grids(obs, order, h_id)
        ordered_result_R = {}
        for result in results_R:
            ordered_result = []
            for x, q, res in zip(x_grid, q_grid, results_R[result]["obs"][0]):
                ordered_result.append(dict(x=x, q=q, res=res))
            ordered_result_R[result] = ordered_result
        return ordered_result_R

    def get_M_results(self, obs, order, h_id):
        mu_list = ["0.5", "1.0", "2.0"]
        filenames_M = [
            obs
            + "_"
            + "M"
            + "_"
            + order
            + "_"
            + h_id
            + "_"
            + heavy_dict[h_id]
            + "_thr="
            + mu
            + "_"
            + pdf
            + "_0"
            for pdf, mu in zip(self.get_restypes(order, h_id, "M")["pdf"], mu_list)
        ]
        results_M = {}
        for filepath in filenames_M:
            with open(self.result_path / (filepath + ".yaml"), encoding="utf-8") as f:
                string = filepath.split(orderstrings[order])[-1].split("_0")[0]
                results_M[string] = yaml.safe_load(f)
        x_grid, q_grid = self.x_q_grids(obs, order, h_id)
        ordered_result_M = {}
        for result in results_M:
            ordered_result = []
            for x, q, res in zip(x_grid, q_grid, results_M[result]["obs"][0]):
                ordered_result.append(dict(x=x, q=q, res=res))
            ordered_result_M[result] = ordered_result
        return ordered_result_M

    def get_M_n3lo_var_results(self, obs, h_id):
        n3lo_cf_variations = ["-1", "1"]
        filenames_M = [
            obs
            + "_"
            + "M"
            + "_"
            + "3"
            + "_"
            + h_id
            + "_"
            + heavy_dict[h_id]
            + "_thr="
            + "1.0"
            + "_"
            + pdf
            + "_"
            + var
            for pdf, var in zip(
                [
                    self.get_restypes("3", h_id, "M")["pdf"][1],
                    self.get_restypes("3", h_id, "M")["pdf"][1],
                ],
                n3lo_cf_variations,
            )
        ]
        results_M = {}
        for filepath, var in zip(filenames_M, n3lo_cf_variations):
            with open(self.result_path / (filepath + ".yaml"), encoding="utf-8") as f:
                results_M[var] = yaml.safe_load(f)
        x_grid, q_grid = self.x_q_grids(obs, "3", h_id)
        ordered_result_M = {}
        for var in results_M:
            ordered_result = []
            for x, q, res in zip(x_grid, q_grid, results_M[var]["obs"][0]):
                ordered_result.append(dict(x=x, q=q, res=res))
            ordered_result_M[var] = ordered_result
        return ordered_result_M

    def plot_single_obs(self, obs, order, h_id):
        n3lo_cf_variations = ["-1", "1"]
        mu_list = ["0.5", "1.0", "2.0"]
        parameters.initialize_theory(True)
        mass = parameters.masses(int(h_id))
        ordered_results_FO, x_grid, _q_grid = self.get_FO_result(obs, order, h_id)
        ordered_result_M = self.get_M_results(obs, order, h_id)
        ordered_result_R = self.get_R_results(obs, order, h_id)
        n3lo_var_FO = {}
        n3lo_var_M = {}
        if order == "3":
            n3lo_var_FO = self.get_FO_n3lo_var_results(obs, h_id)
            n3lo_var_M = self.get_M_n3lo_var_results(obs, h_id)
        diff_x_points = list(set(x_grid))
        for x in diff_x_points:
            plot_name = obs + "_" + order + "_" + h_id
            plot_path = self.plot_dir / (plot_name + "_" + str(x) + ".pdf")
            q_plot = [
                res["q"] for res in ordered_results_FO[mu_list[1]] if res["x"] == x
            ]
            res_plot_FO = [
                res["res"] for res in ordered_results_FO[mu_list[1]] if res["x"] == x
            ]
            res_plot_FO_2mu = [
                res["res"] for res in ordered_results_FO[mu_list[2]] if res["x"] == x
            ]
            res_plot_FO_05mu = [
                res["res"] for res in ordered_results_FO[mu_list[0]] if res["x"] == x
            ]
            res_plot_FO_n3lo_variations = []
            if order == "3":
                res_plot_FO_n3lo_variations.append(
                    [
                        res["res"]
                        for res in n3lo_var_FO[n3lo_cf_variations[0]]
                        if res["x"] == x
                    ]
                )
                res_plot_FO_n3lo_variations.append(
                    [
                        res["res"]
                        for res in n3lo_var_FO[n3lo_cf_variations[1]]
                        if res["x"] == x
                    ]
                )
            sv_FO_coll = [res_plot_FO_05mu, res_plot_FO, res_plot_FO_2mu]
            res_plot_R = {}
            res_plot_M = {}
            shifts = {
                "_mub=05mb": 0.5,
                "": 1.0,
                "_mub=2mb": 2.0,
            }
            corr = {
                "_mub=05mb": 1.12,
                "": 1.0,
                "_mub=2mb": 1.12,
            }
            for sv in ordered_result_R:
                res_plot_tmp = [res for res in ordered_result_R[sv] if res["x"] == x]
                res_plot = [
                    res["res"] if res["q"] >= (shifts[sv] * mass) * corr[sv] else np.nan
                    for res in res_plot_tmp
                ]
                res_plot_R[sv] = res_plot
            for sv, fo_sv in zip(ordered_result_M, sv_FO_coll):
                res_plot_tmp = [res for res in ordered_result_M[sv] if res["x"] == x]
                res_plot = [
                    res["res"] if res["q"] >= (shifts[sv] * mass) * corr[sv] else res_FO
                    for res, res_FO in zip(res_plot_tmp, fo_sv)
                ]
                res_plot_M[sv] = res_plot
            res_plot_M_var = {}
            for var, fo_var in zip(n3lo_var_M, res_plot_FO_n3lo_variations):
                res_plot_tmp = [res for res in n3lo_var_M[var] if res["x"] == x]
                res_plot = [
                    res["res"] if res["q"] >= mass * corr[""] else res_FO
                    for res, res_FO in zip(res_plot_tmp, fo_var)
                ]
                res_plot_M_var[var] = res_plot
            plt.xscale("log")
            plt.plot(
                q_plot,
                res_plot_FO,
                label="FO",
                color="violet",
                linestyle="--",
                linewidth=2.0,
            )
            plt.plot(
                q_plot,
                res_plot_R[list(shifts.keys())[1]],
                label="R",
                color="green",
                linewidth=0.8,
            )
            plt.plot(
                q_plot,
                res_plot_M[list(shifts.keys())[1]],
                label="M",
                color="blue",
                linewidth=1.2,
            )
            to_fill_R = [
                (
                    lambda q: abs(
                        res_plot_R[list(shifts.keys())[0]][q_plot.index(q)]
                        - res_plot_R[list(shifts.keys())[1]][q_plot.index(q)]
                    )
                    if abs(
                        res_plot_R[list(shifts.keys())[0]][q_plot.index(q)]
                        - res_plot_R[list(shifts.keys())[1]][q_plot.index(q)]
                    )
                    > abs(
                        res_plot_R[list(shifts.keys())[2]][q_plot.index(q)]
                        - res_plot_R[list(shifts.keys())[1]][q_plot.index(q)]
                    )
                    else abs(
                        res_plot_R[list(shifts.keys())[2]][q_plot.index(q)]
                        - res_plot_R[list(shifts.keys())[1]][q_plot.index(q)]
                    )
                )(q)
                for q in q_plot
            ]
            to_fill_M = [
                (
                    lambda q: abs(
                        res_plot_M[list(shifts.keys())[0]][q_plot.index(q)]
                        - res_plot_M[list(shifts.keys())[1]][q_plot.index(q)]
                    )
                    if abs(
                        res_plot_M[list(shifts.keys())[0]][q_plot.index(q)]
                        - res_plot_M[list(shifts.keys())[1]][q_plot.index(q)]
                    )
                    > abs(
                        res_plot_M[list(shifts.keys())[2]][q_plot.index(q)]
                        - res_plot_M[list(shifts.keys())[1]][q_plot.index(q)]
                    )
                    else abs(
                        res_plot_M[list(shifts.keys())[2]][q_plot.index(q)]
                        - res_plot_M[list(shifts.keys())[1]][q_plot.index(q)]
                    )
                )(q)
                for q in q_plot
            ]
            to_fill_M_n3lo_var = []
            if order == "3":
                to_fill_M_n3lo_var = [
                    (
                        lambda q: abs(
                            res_plot_M_var["-1"][q_plot.index(q)]
                            - res_plot_M[list(shifts.keys())[1]][q_plot.index(q)]
                        )
                        if abs(
                            res_plot_M_var["-1"][q_plot.index(q)]
                            - res_plot_M[list(shifts.keys())[1]][q_plot.index(q)]
                        )
                        > abs(
                            res_plot_M_var["1"][q_plot.index(q)]
                            - res_plot_M[list(shifts.keys())[1]][q_plot.index(q)]
                        )
                        else abs(
                            res_plot_M_var["1"][q_plot.index(q)]
                            - res_plot_M[list(shifts.keys())[1]][q_plot.index(q)]
                        )
                    )(q)
                    for q in q_plot
                ]
            to_fill_FO_n3lo_var = []
            if order == "3":
                to_fill_FO_n3lo_var = [
                    (
                        lambda q: abs(
                            res_plot_FO_n3lo_variations[0][q_plot.index(q)]
                            - res_plot_FO[q_plot.index(q)]
                        )
                        if abs(
                            res_plot_FO_n3lo_variations[0][q_plot.index(q)]
                            - res_plot_FO[q_plot.index(q)]
                        )
                        > abs(
                            res_plot_FO_n3lo_variations[1][q_plot.index(q)]
                            - res_plot_FO[q_plot.index(q)]
                        )
                        else abs(
                            res_plot_FO_n3lo_variations[1][q_plot.index(q)]
                            - res_plot_FO[q_plot.index(q)]
                        )
                    )(q)
                    for q in q_plot
                ]
            plt.fill_between(
                q_plot,
                np.array(res_plot_R[list(shifts.keys())[1]]) + np.array(to_fill_R),
                np.array(res_plot_R[list(shifts.keys())[1]]) - np.array(to_fill_R),
                color="green",
                label="scale_unc",
                alpha=0.25,
            )
            plt.fill_between(
                q_plot,
                np.array(res_plot_M[list(shifts.keys())[1]]) + np.array(to_fill_M),
                np.array(res_plot_M[list(shifts.keys())[1]]) - np.array(to_fill_M),
                color="blue",
                label="scale_unc",
                alpha=0.25,
            )
            if order == "3":
                plt.fill_between(
                    q_plot,
                    np.array(res_plot_M[list(shifts.keys())[1]])
                    + np.array(to_fill_M_n3lo_var),
                    np.array(res_plot_M[list(shifts.keys())[1]])
                    - np.array(to_fill_M_n3lo_var),
                    color="lightseagreen",
                    label="cf_unc",
                    alpha=0.25,
                )
                plt.fill_between(
                    q_plot,
                    np.array(res_plot_FO) + np.array(to_fill_FO_n3lo_var),
                    np.array(res_plot_FO) - np.array(to_fill_FO_n3lo_var),
                    color="darkorchid",
                    label="cf_unc",
                    alpha=0.25,
                    linestyle="--",
                )
            plt.xlabel("Q[GeV]")
            plt.ylabel("x" + obs)
            plt.legend()
            plt.grid(alpha=0.75)
            plt.savefig(plot_path)
            plt.close()

    def plot_fonll_noerr(self, obs, order, h_id):
        parameters.initialize_theory(True)
        mass = parameters.masses(int(h_id))
        ordered_result_FO, x_grid, _q_grid = self.get_FO_result(obs, order, h_id)
        ordered_result_M = self.get_M_results(obs, order, h_id)[""]
        ordered_result_R = self.get_R_results(obs, order, h_id)[""]
        diff_x_points = list(set(x_grid))
        for x in diff_x_points:
            plot_name = obs + "_" + order + "_" + h_id
            plot_path = self.plot_dir / (plot_name + "_" + str(x) + ".pdf")
            q_plot = [res["q"] for res in ordered_result_FO if res["x"] == x]
            res_plot_FO = [res["res"] for res in ordered_result_FO if res["x"] == x]
            res_plot_tmp = [res for res in ordered_result_R if res["x"] == x]
            res_plot = [
                res["res"] if res["q"] > mass else np.nan for res in res_plot_tmp
            ]
            res_plot_R = res_plot
            res_plot_tmp = [res for res in ordered_result_M if res["x"] == x]
            res_plot = [
                res["res"] if res["q"] > mass else np.nan for res in res_plot_tmp
            ]
            res_plot_M = res_plot
            plt.plot(
                q_plot,
                res_plot_FO,
                label="Massive",
                color="violet",
                linestyle="--",
                linewidth=3.5,
            )
            plt.plot(
                q_plot,
                res_plot_R,
                label="ZM",
                color="green",
                linewidth=0.8,
            )
            plt.plot(
                q_plot,
                res_plot_M,
                label="FONLL",
                color="blue",
                linewidth=2.2,
            )
            plt.xscale("log")
            plt.xlabel("Q[GeV]")
            plt.ylabel("x" + obs)
            plt.legend()
            plt.grid(alpha=0.75)
            plt.savefig(plot_path)
            plt.close()

    def plot_fonll_order_comparison(self, obs, h_id):
        parameters.initialize_theory(True)
        mass = parameters.masses(int(h_id))
        ordered_result_FO_NNLO, x_grid, _q_grid = self.get_FO_result(obs, "NNLO", h_id)
        ordered_result_M_NNLO = self.get_M_results(obs, "NNLO", h_id)[""]
        ordered_result_M_N3LO = self.get_M_results(obs, "N3LO", h_id)[""]
        diff_x_points = list(set(x_grid))
        for x in diff_x_points:
            plot_name = obs + "_comporders_" + h_id
            plot_path = self.plot_dir / (plot_name + "_" + str(x) + ".pdf")
            q_plot = [res["q"] for res in ordered_result_FO_NNLO if res["x"] == x]
            res_plot_tmp = [res for res in ordered_result_M_NNLO if res["x"] == x]
            res_plot = [res["res"] if res["q"] > mass else 0.0 for res in res_plot_tmp]
            res_plot_M_NNLO = res_plot
            res_plot_tmp = [res for res in ordered_result_M_N3LO if res["x"] == x]
            res_plot = [res["res"] if res["q"] > mass else 0.0 for res in res_plot_tmp]
            res_plot_M_N3LO = res_plot
            plt.plot(
                q_plot,
                res_plot_M_NNLO,
                label="NNLO",
                color="violet",
                linewidth=2.0,
            )
            plt.plot(
                q_plot,
                res_plot_M_N3LO,
                label="N3LO",
                color="blue",
                linewidth=2.5,
            )
            plt.xscale("log")
            plt.xlabel("Q[GeV]")
            plt.ylabel("x" + obs)
            plt.legend()
            plt.grid(alpha=0.75)
            plt.savefig(plot_path)
            plt.close()



import lhapdf
import numpy as np
from scipy import integrate

from ..parameters import charges


def mkPDF(pdf_name, order):
    lhapdf.setVerbosity(0)
    Mypdf = None
    if isinstance(pdf_name, list):
        Mypdf = lhapdf.mkPDF(pdf_name[order - 1], 0)
    elif isinstance(pdf_name, str):
        Mypdf = lhapdf.mkPDF(pdf_name, 0)
    return Mypdf


def isospin_roation(A, Z, nf):
    """Return a roation matrix for light flavors."""
    rot = np.eye(nf)
    rot[0, 0] = rot[1, 1] = 1 / A * Z
    rot[1, 0] = rot[0, 1] = 1 / A * (A - Z)
    return rot


def apply_isospin_roation(pdf, x, Q, light_f, target_dict):
    """Apply the isospin roation for a given pdf."""
    isospin_conj = isospin_roation(target_dict["A"], target_dict["Z"], len(light_f))
    light_q = np.array([pdf.xfxQ2(nl, x, Q * Q) for nl in light_f])
    light_qbar = np.array([pdf.xfxQ2(-nl, x, Q * Q) for nl in light_f])
    return isospin_conj @ light_q + light_qbar


def non_singlet_pdf(pdf, x, Q, nf, target_dict):
    """Return the `NonSinglet` flavor combination."""
    light_f = [1, 2, 3]
    if nf >= 4:
        light_f.append(4)
    if nf >= 5:
        light_f.append(5)
    light_pdfs = apply_isospin_roation(pdf, x, Q, light_f, target_dict)
    e2_q = np.array([charges(nl) ** 2 for nl in light_f])
    return np.sum(e2_q * light_pdfs)


def singlet_pdf(pdf, x, Q, nf, target_dict):
    """Return the `Singlet` flavor combination."""
    light_f = [1, 2, 3]
    if nf >= 4:
        light_f.append(4)
    if nf >= 5:
        light_f.append(5)
    light_pdfs = apply_isospin_roation(pdf, x, Q, light_f, target_dict)
    return np.sum(light_pdfs)


def PDFConvolute_light(func1, pdf, x, Q, p1, nf, target_dict):
    result, _ = integrate.quad(
        lambda z: func1(z, Q, p1, nf) * non_singlet_pdf(pdf, x / z, Q, nf, target_dict),
        x,
        1.0,
        epsabs=1e-12,
        epsrel=1e-6,
        limit=200,
        points=(x, 1.0),
    )
    return result


def PDFConvolute_light_singlet(func1, pdf, x, Q, p1, nf, target_dict):
    result, _ = integrate.quad(
        lambda z: func1(z, Q, p1, nf) * singlet_pdf(pdf, x / z, Q, nf, target_dict),
        x,
        1.0,
        epsabs=1e-12,
        epsrel=1e-6,
        limit=200,
        points=(x, 1.0),
    )
    return result


def PDFConvolute_light_plus(func1, pdf, x, Q, p1, nf, target_dict):
    result, _ = integrate.quad(
        lambda z: func1(z, Q, p1, nf)
        * (
            non_singlet_pdf(pdf, x / z, Q, nf, target_dict)
            - non_singlet_pdf(pdf, x, Q, nf, target_dict)
        ),
        x,
        1.0,
        epsabs=1e-12,
        epsrel=1e-6,
        limit=200,
        points=(x, 1.0),
    )
    return result



# These are utility functions to read txt grids.
import numpy as np
from multiprocess import Pool

from . import Initialize as Ini
from .logging import console
from .parameters import charges, default_masses


def read1D(path_to_file):
    """
    Read a space-separated txt file and return a 1-Dimensional list of the values

    Parameters:
        path_to_file : str
            file to open
    Returns:
            : list
        list of values
    """
    for line in open(path_to_file):
        listWord = line.split(" ")
    mylist = [float(item) for item in listWord]
    return mylist


def readND(path_to_file):
    """
    Read a space-separated txt file and return a N-Dimensional list of the values

    Input:
        path_to_file : str
            file to open
    Returns:
            : list
        list of values
    """
    mylist = []
    for line in open(path_to_file):
        listWord = line.split(" ")
        mylist.append(listWord)
    list = [
        [
            float(item)
            if item != "-nan" and item != "nan"
            else float(mylist[a][mylist[a].index(item) - 1])
            for item in mylist[a]
        ]
        for a in range(len(mylist))
    ]
    return list


class Construct_Grid:
    def __init__(self, func, h_id, path, grid_type, n_pools=8):
        self.func = func
        self.mass = default_masses(h_id)
        self.path = path
        self.xgrid = Ini.ZList
        self.qgrid = Ini.QList
        self.n_pools = n_pools
        self.nf = h_id
        self.e_h = charges(h_id)
        self.grid_type = grid_type
        if self.grid_type == "tilde":
            self.xgrid = self.xgrid[:-1]

    def construct_single_x(self, z):
        z_func_values = []
        p = []
        i = self.xgrid.index(z)
        console.log(f"[green]Computing x = {z},  {i}/{len(self.xgrid)}")
        for q in self.qgrid:
            p = [self.mass, q, self.e_h]
            if self.grid_type == "matching":
                z_func_values.append(self.func(z, p, self.nf, use_analytic=True))
            elif self.grid_type == "tilde":
                z_func_values.append(self.func(z, q, p, self.nf, use_analytic=True))
        return z_func_values

    def run(self):
        args = (self.construct_single_x, self.xgrid)
        with Pool(self.n_pools) as pool:
            result = pool.map(*args)

        func_values = []
        for res in result:
            func_values.append(res)
        console.log(f"Computation finished, saving to {self.path}")
        np.savetxt(self.path, func_values)
        return func_values



# this contains the FO massive coefficients functions.

import LeProHQ
import numpy as np
from eko.constants import TR
from scipy.integrate import quad

from . import Initialize


# F2
def Cb_2_m_reg(z, Q, p, _nf):
    Q2 = Q * Q
    m_b = p[0]
    eps = m_b * m_b / Q2
    thre = 4.0 * eps * z / (1 - z)
    if thre > 1.0:
        return 0
    e_h = p[-1]
    xi = 1 / eps
    eta = xi / 4.0 * (1.0 / z - 1.0) - 1.0
    eta = min(eta, 1e5)
    FHprefactor = Q2 / (np.pi * m_b**2) * e_h**2
    return FHprefactor / z * (4.0 * np.pi) ** 2 * LeProHQ.dq1("F2", "VV", xi, eta)


def Cb_2_m_loc(_z, Q, p, _nf):
    l = quad(
        lambda x: Cb_2_m_reg(x, Q, p, _nf),
        0.0,
        1.0,
        points=(0.0, 1.0),
    )
    return -l[0]


def Cg_1_m_reg(z, Q, p, _nf):
    Q2 = Q * Q
    m_b = p[0]
    eps = m_b * m_b / Q2
    thre = 4.0 * eps * z / (1 - z)
    e_h = p[-1]
    if thre > 1.0:
        return 0
    v = np.sqrt(1 - thre)
    return (
        4
        * TR
        * e_h
        * e_h
        * (
            v * (8 * z * (1 - z) - 1 - 4 * z * (1 - z) * eps)
            + np.log((1 + v) / (1 - v))
            * (z * z + (1 - z) ** 2 + 4 * z * eps * (1 - 3 * z) - 8 * z * z * eps * eps)
        )
    )


def Cg_2_m_reg(z, Q, p, _nf):
    Q2 = Q * Q
    m_b = p[0]
    eps = m_b**2 / Q2
    thre = 4.0 * eps * z / (1 - z)
    if thre > 1.0:
        return 0
    e_h = p[-1]
    xi = 1 / eps
    eta = xi / 4.0 * (1.0 / z - 1.0) - 1.0
    FHprefactor = Q2 / (np.pi * m_b**2) * e_h**2
    if xi > 2499.9999999999995:
        # FH grids are not defined above this
        return 0.0
    return (
        FHprefactor
        / z
        * (4.0 * np.pi) ** 2
        * (
            LeProHQ.cg1("F2", "VV", xi, eta)
            + LeProHQ.cgBar1("F2", "VV", xi, eta) * np.log(xi)
        )
    )


def Cg_3_m_reg(z, Q, p, nf):
    e_h = p[-1]
    Q2 = Q * Q
    m_b = p[0]
    eps = m_b * m_b / Q2
    thre = 4.0 * eps * z / (1 - z)
    if thre > 1.0:
        return 0.0
    return e_h**2 * Initialize.Cg3m[nf - 4](z, Q)[0]


def Cq_2_m_reg(z, Q, p, _nf):
    Q2 = Q * Q
    m_b = p[0]
    eps = m_b * m_b / Q2
    thre = 4.0 * eps * z / (1 - z)
    if thre > 1.0:
        return 0
    e_h = p[-1]
    xi = 1 / eps
    eta = xi / 4.0 * (1.0 / z - 1.0) - 1.0
    FHprefactor = Q2 / (np.pi * m_b**2) * e_h**2
    return (
        FHprefactor
        / z
        * (4.0 * np.pi) ** 2
        * (
            LeProHQ.cq1("F2", "VV", xi, eta)
            + LeProHQ.cqBarF1("F2", "VV", xi, eta) * np.log(xi)
        )
    )


def Cq_3_m_reg(z, Q, p, nf):
    e_h = p[-1]
    Q2 = Q * Q
    m_b = p[0]
    eps = m_b * m_b / Q2
    thre = 4.0 * eps * z / (1 - z)
    if thre > 1.0:
        return 0.0
    return e_h**2 * Initialize.Cq3m[nf - 4](z, Q)[0]


# FL
def CLb_2_m_reg(z, Q, p, _nf):
    Q2 = Q * Q
    m_b = p[0]
    eps = m_b * m_b / Q2
    thre = 4.0 * eps * z / (1 - z)
    if thre > 1.0:
        return 0
    e_h = p[-1]
    xi = 1 / eps
    eta = xi / 4.0 * (1.0 / z - 1.0) - 1.0
    FHprefactor = Q2 / (np.pi * m_b**2) * e_h**2
    return FHprefactor / z * (4.0 * np.pi) ** 2 * LeProHQ.dq1("FL", "VV", xi, eta)


def CLg_1_m_reg(z, Q, p, _nf):
    Q2 = Q * Q
    m_b = p[0]
    eps = m_b * m_b / Q2
    z2 = z * z
    thre = 4.0 * eps * z / (1 - z)
    e_h = p[-1]
    if thre > 1.0:
        return 0
    v = np.sqrt(1 - thre)
    return (
        4
        * TR
        * e_h
        * e_h
        * (-8 * eps * z2 * np.log((1 + v) / (1 - v)) + 4 * v * z * (1 - z))
    )


def CLg_2_m_reg(z, Q, p, _nf):
    Q2 = Q * Q
    m_b = p[0]
    eps = m_b * m_b / Q2
    thre = 4.0 * eps * z / (1 - z)
    if thre > 1.0:
        return 0
    e_h = p[-1]
    xi = 1 / eps
    eta = xi / 4.0 * (1.0 / z - 1.0) - 1.0
    FHprefactor = Q2 / (np.pi * m_b**2) * e_h**2
    if xi > 2499.9999999999995:
        # FH grids are not defined above this
        return 0.0
    return (
        FHprefactor
        / z
        * (4.0 * np.pi) ** 2
        * (
            LeProHQ.cg1("FL", "VV", xi, eta)
            + LeProHQ.cgBar1("FL", "VV", xi, eta) * np.log(xi)
        )
    )


def CLg_3_m_reg(z, Q, p, nf):
    e_h = p[-1]
    Q2 = Q * Q
    m_b = p[0]
    eps = m_b * m_b / Q2
    thre = 4.0 * eps * z / (1 - z)
    if thre > 1.0:
        return 0
    return e_h**2 * Initialize.CLg3m[nf - 4](z, Q)[0]


def CLq_2_m_reg(z, Q, p, _nf):
    Q2 = Q * Q
    m_b = p[0]
    eps = m_b * m_b / Q2
    thre = 4.0 * eps * z / (1 - z)
    if thre > 1.0:
        return 0
    e_h = p[-1]
    xi = 1 / eps
    eta = xi / 4.0 * (1.0 / z - 1.0) - 1.0
    FHprefactor = Q2 / (np.pi * m_b**2) * e_h**2
    return (
        FHprefactor
        / z
        * (4.0 * np.pi) ** 2
        * (
            LeProHQ.cq1("FL", "VV", xi, eta)
            + LeProHQ.cqBarF1("FL", "VV", xi, eta) * np.log(xi)
        )
    )


def CLq_3_m_reg(z, Q, p, nf):
    e_h = p[-1]
    Q2 = Q * Q
    m_b = p[0]
    eps = m_b * m_b / Q2
    thre = 4.0 * eps * z / (1 - z)
    if thre > 1.0:
        return 0.0
    return e_h**2 * Initialize.CLq3m[nf - 4](z, Q)[0]



"""Script to produce the k-factors."""

import copy
from datetime import date

import lhapdf
import numpy as np
import pandas as pd
import yadism

from . import configs
from .io import OperatorParameters, TheoryParameters
from .logging import console, df_to_table
from .runner import Runner


class KfactorRunner:
    def __init__(
        self,
        t_card_name,
        dataset_name,
        pdf_name,
        use_yadism,
        fonll_incomplete,
        cfg_path,
    ):
        cfg = configs.load(cfg_path)
        cfg = configs.defaults(cfg)

        self.theory = TheoryParameters.load_card(cfg, t_card_name)
        self.observable = OperatorParameters.load_card(cfg, dataset_name, pdf_name)

        self.pdf_name = pdf_name
        self.use_yadism = use_yadism
        self.fonll_incomplete = fonll_incomplete
        self.result_path = cfg["paths"]["results"] / t_card_name
        self.dataset_name = dataset_name
        self._results = None
        self.config_path = cfg_path

    def run_yadism(self):
        output = yadism.run_yadism(
            self.theory.yadism_like(), self.observable.yadism_like()
        )
        yad_pred = output.apply_pdf(lhapdf.mkPDF(self.pdf_name))
        return yad_pred

    def run_dis_tp(self, n_cores):
        runner = Runner(self.observable, self.theory, self.config_path)
        runner.compute(n_cores)
        return copy.deepcopy(runner.results)

    def compute(self, n_cores):
        mumerator_log = self.run_dis_tp(n_cores)
        # over Yadism N3LO
        if self.use_yadism:
            denominator_log = self.run_yadism()
        else:
            # over FONLL incomplete N3LO
            if self.fonll_incomplete:
                self._update_observables_to_incomplete()
            # over NNLO
            else:
                self.theory.order -= 1
            denominator_log = self.run_dis_tp(n_cores)
        logs_df = self._log(
            mumerator_log, denominator_log, self.use_yadism, self.fonll_incomplete
        )
        self._results = self.build_kfactor(logs_df)
        console.log(df_to_table(self._results, self.dataset_name))

    def _update_observables_to_incomplete(self):
        for obs in self.observable.obs:
            obs.restype = f"{obs.restype}_incomplete"

    @staticmethod
    def _log(num, den, use_yadism, fonll_incomplete):
        # loop on SF
        for obs in den:
            my_obs = obs.split("_")[0]
            if fonll_incomplete:
                den_df = den[my_obs].rename(columns={"result": "N3LO_incomplete"})
                num_df = num[my_obs].rename(columns={"result": "N3LO"})
            elif use_yadism:
                den_df = pd.DataFrame(den[obs]).rename(columns={"result": "yadism"})
                num_df = num[my_obs].rename(columns={"result": "dis_tp"})
            else:
                den_df = den[my_obs].rename(columns={"result": "NNLO"})
                num_df = num[my_obs].rename(columns={"result": "N3LO"})
            log_df = pd.concat([den_df, num_df], axis=1).T.drop_duplicates().T

            # construct some nice log table
            log_df.drop("y", axis=1, inplace=True)
            if use_yadism:
                log_df.drop("q", axis=1, inplace=True)
                log_df.drop("error", axis=1, inplace=True)
            else:
                log_df["Q2"] = log_df.q**2
                log_df.drop("q", axis=1, inplace=True)
        return log_df

    def build_kfactor(self, log_df):
        if self.fonll_incomplete:
            log_df["k-factor"] = log_df.N3LO / log_df.N3LO_incomplete
        elif self.use_yadism:
            log_df["k-factor"] = log_df.dis_tp / log_df.yadism
        else:
            log_df["k-factor"] = log_df.N3LO / log_df.NNLO
        log_df["kf_error"] = 0.0
        return log_df

    @staticmethod
    def build_kfactor_with_error(reuslts):
        kfs = []
        for res in reuslts:
            kfs.append(res["k-factor"])
        kfs = np.array(kfs)
        return pd.DataFrame({"k-factor": kfs.mean(axis=0), "kf_error": kfs.std(axis=0)})

    def save_results(self, author, th_input):
        if self.fonll_incomplete:
            k_fatctor_type = (
                "FONLL@N3LO DIS_TP / (FONLL@NNLO + ZM-VFNS@N3LO_only) DIS_TP"
            )
        elif self.use_yadism:
            k_fatctor_type = (
                "FONLL@N3LO DIS_TP / (FONLL@NNLO + ZM-VFNS@N3LO_only) Yadism"
            )
        else:
            k_fatctor_type = "FONLL@N3LO DIS_TP / FONLL@NNLO DIS_TP"
        intro = [
            "********************************************************************************\n",
            f"SetName: {self.dataset_name}\n",
            f'Author: {author.replace("_", " ")}\n',
            f"Date: {date.today()}\n",
            "CodesUsed: https://github.com/andreab1997/DIS_TP\n",
            f"TheoryInput: {th_input}\n",
            f"PDFset: {self.pdf_name}\n",
            f"Warnings: {k_fatctor_type}\n"
            "********************************************************************************\n",
        ]
        self.result_path.mkdir(exist_ok=True)
        res_path = self.result_path / f"CF_QCD_{self.dataset_name}.dat"
        console.log(f"[green]Saving the k-factors in: {res_path}")
        with open(res_path, "w", encoding="utf-8") as f:
            f.writelines(intro)
            f.writelines(
                [
                    f"{k:4f}   {e:4f}\n"
                    for k, e in zip(
                        self._results["k-factor"], self._results["kf_error"]
                    )
                ]
            )



import pathlib

import click

from .. import Initialize
from .. import MatchingFunc as mcf
from .. import ReadTxt
from .. import TildeCoeffFunc as tcf
from .. import parameters
from .base import command, root_path


@command.group("grids")
def subcommand():
    """Generate grids"""


heavy_id = click.argument("h_id", type=int)

pto = click.argument("PTO", type=int)

flavor_entry = click.argument(
    "flavor_entry",
    type=str,
)

n_cores = click.option(
    "-n",
    "--n_cores",
    type=int,
    default=4,
    required=False,
    help="Number of cores",
)

dest_path = click.option(
    "-d",
    "--dest_path",
    type=pathlib.Path,
    default=root_path,
    required=False,
    help="destination path",
)


@subcommand.command("matching")
@heavy_id
@flavor_entry
@pto
@n_cores
@dest_path
def generate_matching_grids(
    h_id: int, flavor_entry: str, pto: int, n_cores: int, dest_path: pathlib.Path
):
    """Construct matching grids."""

    if pto == 3:
        if flavor_entry == "bq":
            func = mcf.Mbq_3_reg
        elif flavor_entry == "bg":
            func = mcf.Mbg_3_reg
    path = dest_path / f"M{flavor_entry}_{pto}/M{flavor_entry}{pto}_nf{h_id}.txt"
    Initialize.Initialize_all(h_id)
    parameters.initialize_theory(use_grids=True)
    obj = ReadTxt.Construct_Grid(
        func, h_id=h_id, path=path, grid_type="matching", n_pools=n_cores
    )
    obj.run()


@subcommand.command("tilde")
@heavy_id
@flavor_entry
@pto
@n_cores
@click.option(
    "-v",
    "--n3lo_variation",
    type=int,
    default=0,
    required=False,
    help="N3LO coeffieicnt variation",
)
@dest_path
def generate_matching_grids(
    h_id: int,
    flavor_entry: str,
    pto: int,
    n_cores: int,
    n3lo_variation: int,
    dest_path: pathlib.Path,
):
    """Construct tilde grids."""
    flavor = flavor_entry[-1] if "2" in flavor_entry else flavor_entry

    if pto == 3:
        if flavor_entry == "2q":
            func = tcf.Cq_3_til_reg
        elif flavor_entry == "2g":
            func = tcf.Cg_3_til_reg
        elif flavor_entry == "Lq":
            func = tcf.CLq_3_til_reg
        elif flavor_entry == "Lg":
            func = tcf.CLg_3_til_reg

    path = (
        dest_path
        / f"C{flavor}_{pto}_til/C{flavor}{pto}til_nf{h_id}_var{n3lo_variation}.txt"
    )
    Initialize.Initialize_all(h_id, n3lo_variation)
    parameters.initialize_theory(use_grids=True)
    obj = ReadTxt.Construct_Grid(
        func, h_id=h_id, path=path, grid_type="tilde", n_pools=n_cores
    )
    obj.run()



import pathlib

import click

root_path = pathlib.Path.cwd().absolute()
"""Default destination for generated files"""


@click.group(context_settings={"help_option_names": ["-h", "--help"]})
def command():
    pass



import functools

import numpy as np
from multiprocess import Pool

from dis_tp import Initialize as Ini

from . import configs, io
from .logging import console
from .parameters import initialize_theory
from .structure_functions import f2, fl

mapfunc = {
    "F2": {
        "R": [f2.F2_R],
        "M": [f2.F2_M],
        "FO": [f2.F2_FO],
        "light": [f2.F2_Light],
        "total": [f2.F2_Total],
        "FONLL": [f2.F2_FONLL],
        "FONLL_incomplete": [f2.F2_FONLL_incomplete],
        "total_incomplete": [f2.F2_Total_incomplete],
    },
    "FL": {
        "R": [fl.FL_R],
        "M": [fl.FL_M],
        "FO": [fl.FL_FO],
        "light": [fl.FL_Light],
        "total": [fl.FL_Total],
        "FONLL": [fl.FL_FONLL],
        "FONLL_incomplete": [fl.FL_FONLL_incomplete],
        "total_incomplete": [fl.FL_Total_incomplete],
    },
    "XSHERANCAVG": {
        "R": [f2.F2_R, fl.FL_R],
        "M": [f2.F2_M, fl.FL_M],
        "FO": [f2.F2_FO, fl.FL_FO],
        "FONLL": [f2.F2_FONLL, fl.FL_FONLL],
        "light": [f2.F2_Light, fl.FL_Light],
        "total": [f2.F2_Total, fl.FL_Total],
        "FONLL_incomplete": [f2.F2_FONLL_incomplete, fl.FL_FONLL_incomplete],
        "total_incomplete": [f2.F2_Total_incomplete, fl.FL_Total_incomplete],
    },
    # NOTE: for the moment this coincide with the averaged xs
    # since here we don't provide F3
    "XSHERANC": {
        "R": [f2.F2_R, fl.FL_R],
        "M": [f2.F2_M, fl.FL_M],
        "FO": [f2.F2_FO, fl.FL_FO],
        "FONLL": [f2.F2_FONLL, fl.FL_FONLL],
        "light": [f2.F2_Light, fl.FL_Light],
        "total": [f2.F2_Total, fl.FL_Total],
        "FONLL_incomplete": [f2.F2_FONLL_incomplete, fl.FL_FONLL_incomplete],
        "total_incomplete": [f2.F2_Total_incomplete, fl.FL_Total_incomplete],
    },
}

map_heavyness = {"charm": 4, "bottom": 5, "light": None, "total": None}


# TODO: rename External to be grids
class Runner:
    def __init__(self, o_card, t_card, config_path=None) -> None:
        cfg = configs.load(config_path)
        cfg = configs.defaults(cfg)
        dest_path = cfg["paths"]["results"]
        if isinstance(t_card, io.TheoryParameters):
            th_obj = t_card
        else:
            th_obj = io.TheoryParameters.load_card(cfg, t_card)
        if isinstance(o_card, io.OperatorParameters):
            obs_obj = o_card
        else:
            obs_obj = io.OperatorParameters.load_card(cfg, o_card)

        self.runparameters = io.RunParameters(th_obj, obs_obj, dest_path)
        self.o_par = self.runparameters.operator_parameters()
        self.t_par = self.runparameters.theory_parameters()

        initialize_theory(
            th_obj.grids, th_obj.masses, th_obj.strong_coupling, th_obj.thr_atlas
        )
        self.partial_sf = None

    @staticmethod
    def compute_xs(ob, sfs):
        """Assembly the XS if needed according to 'XSHERANCAVG'"""
        yp = 1.0 + (1.0 - ob.y_grid) ** 2
        yL = ob.y_grid**2
        xs = sfs[0] - yL / yp * sfs[1]
        return xs

    def compute_sf(self, kins):
        x, q = kins
        # console.log(f"x={x}, Q={q}")
        return float(self.partial_sf(x=x, Q=q))

    def compute(self, n_cores):
        # loop on observables
        for ob in self.o_par.obs:

            hid = map_heavyness[ob.heavyness]
            if ob.heavyness != "light" and self.t_par.order == 3 and self.t_par.grids:
                Ini.Initialize_all(n3lo_variation=self.t_par.n3lo_variation)

            func_to_call = mapfunc[ob.name][ob.restype]
            thisob_res = []
            # loop on SF
            for func in func_to_call:
                self.partial_sf = functools.partial(
                    func,
                    order=self.t_par.order,
                    meth=self.t_par.fns,
                    pdf=ob.pdf,
                    h_id=hid,
                    target_dict=self.o_par.target_dict,
                )
                console.log(
                    f"[green]Computing {func.__name__} @ order: {self.t_par.order} ..."
                )
                args = (self.compute_sf, zip(ob.x_grid, ob.q_grid))
                if n_cores == 1:
                    sf_map = map(*args)
                    sf_res = np.array([res for res in sf_map])
                else:
                    with Pool(n_cores) as pool:
                        sf_res = pool.map(*args)

                thisob_res.append(sf_res)
            thisob_res = np.array(thisob_res)
            if ob.name in ["XSHERANC", "XSHERANCAVG"]:
                thisob_res = self.compute_xs(ob, thisob_res)
            self.runparameters.results[ob] = thisob_res

    @property
    def results(self):
        """Return computed results as dictionary"""
        log = {}
        for ob, vals in self.runparameters.results.items():
            df = ob.kinematics
            df["result"] = vals.T
            log[ob.name] = df
        return log

    def save_results(self):
        self.runparameters.dump_results()
